# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
import os
import platform
import warnings

from pyglet import com, image
from pyglet.util import debug_print
from pyglet.libs.win32 import _kernel32 as kernel32
from pyglet.libs.win32 import _ole32 as ole32
from pyglet.libs.win32.constants import *
from pyglet.libs.win32.types import *
from pyglet.media import Source, MediaDecodeException
from pyglet.media.codecs import AudioFormat, AudioData, VideoFormat, MediaDecoder, StaticSource

_debug = debug_print('debug_media')

try:
    mfreadwrite = 'mfreadwrite'
    mfplat = 'mfplat'

    # System32 and SysWOW64 folders are opposite perception in Windows x64.
    # System32 = x64 dll's | SysWOW64 = x86 dlls
    # By default ctypes only seems to look in system32 regardless of Python architecture, which has x64 dlls.
    if platform.architecture()[0] == '32bit':
        if platform.machine().endswith('64'):  # Machine is 64 bit, Python is 32 bit.
            mfreadwrite = os.path.join(os.environ['WINDIR'], 'SysWOW64', 'mfreadwrite.dll')
            mfplat = os.path.join(os.environ['WINDIR'], 'SysWOW64', 'mfplat.dll')

    mfreadwrite_lib = ctypes.windll.LoadLibrary(mfreadwrite)
    mfplat_lib = ctypes.windll.LoadLibrary(mfplat)
except OSError:
    # Doesn't exist? Should stop import of library.
    raise ImportError('Could not load WMF library.')

MF_SOURCE_READERF_ERROR = 0x00000001
MF_SOURCE_READERF_ENDOFSTREAM = 0x00000002
MF_SOURCE_READERF_NEWSTREAM = 0x00000004
MF_SOURCE_READERF_NATIVEMEDIATYPECHANGED = 0x00000010
MF_SOURCE_READERF_CURRENTMEDIATYPECHANGED = 0x00000020
MF_SOURCE_READERF_STREAMTICK = 0x00000100

# Audio attributes
MF_LOW_LATENCY = com.GUID(0x9c27891a, 0xed7a, 0x40e1, 0x88, 0xe8, 0xb2, 0x27, 0x27, 0xa0, 0x24, 0xee)

# Audio information
MF_MT_ALL_SAMPLES_INDEPENDENT = com.GUID(0xc9173739, 0x5e56, 0x461c, 0xb7, 0x13, 0x46, 0xfb, 0x99, 0x5c, 0xb9, 0x5f)
MF_MT_FIXED_SIZE_SAMPLES = com.GUID(0xb8ebefaf, 0xb718, 0x4e04, 0xb0, 0xa9, 0x11, 0x67, 0x75, 0xe3, 0x32, 0x1b)
MF_MT_SAMPLE_SIZE = com.GUID(0xdad3ab78, 0x1990, 0x408b, 0xbc, 0xe2, 0xeb, 0xa6, 0x73, 0xda, 0xcc, 0x10)
MF_MT_COMPRESSED = com.GUID(0x3afd0cee, 0x18f2, 0x4ba5, 0xa1, 0x10, 0x8b, 0xea, 0x50, 0x2e, 0x1f, 0x92)
MF_MT_WRAPPED_TYPE = com.GUID(0x4d3f7b23, 0xd02f, 0x4e6c, 0x9b, 0xee, 0xe4, 0xbf, 0x2c, 0x6c, 0x69, 0x5d)
MF_MT_AUDIO_NUM_CHANNELS = com.GUID(0x37e48bf5, 0x645e, 0x4c5b, 0x89, 0xde, 0xad, 0xa9, 0xe2, 0x9b, 0x69, 0x6a)
MF_MT_AUDIO_SAMPLES_PER_SECOND = com.GUID(0x5faeeae7, 0x0290, 0x4c31, 0x9e, 0x8a, 0xc5, 0x34, 0xf6, 0x8d, 0x9d, 0xba)
MF_MT_AUDIO_FLOAT_SAMPLES_PER_SECOND = com.GUID(0xfb3b724a, 0xcfb5, 0x4319, 0xae, 0xfe, 0x6e, 0x42, 0xb2, 0x40, 0x61, 0x32)
MF_MT_AUDIO_AVG_BYTES_PER_SECOND = com.GUID(0x1aab75c8, 0xcfef, 0x451c, 0xab, 0x95, 0xac, 0x03, 0x4b, 0x8e, 0x17, 0x31)
MF_MT_AUDIO_BLOCK_ALIGNMENT = com.GUID(0x322de230, 0x9eeb, 0x43bd, 0xab, 0x7a, 0xff, 0x41, 0x22, 0x51, 0x54, 0x1d)
MF_MT_AUDIO_BITS_PER_SAMPLE = com.GUID(0xf2deb57f, 0x40fa, 0x4764, 0xaa, 0x33, 0xed, 0x4f, 0x2d, 0x1f, 0xf6, 0x69)
MF_MT_AUDIO_VALID_BITS_PER_SAMPLE = com.GUID(0xd9bf8d6a, 0x9530, 0x4b7c, 0x9d, 0xdf, 0xff, 0x6f, 0xd5, 0x8b, 0xbd, 0x06)
MF_MT_AUDIO_SAMPLES_PER_BLOCK = com.GUID(0xaab15aac, 0xe13a, 0x4995, 0x92, 0x22, 0x50, 0x1e, 0xa1, 0x5c, 0x68, 0x77)
MF_MT_AUDIO_CHANNEL_MASK = com.GUID(0x55fb5765, 0x644a, 0x4caf, 0x84, 0x79, 0x93, 0x89, 0x83, 0xbb, 0x15, 0x88)
MF_PD_DURATION = com.GUID(0x6c990d33, 0xbb8e, 0x477a, 0x85, 0x98, 0xd, 0x5d, 0x96, 0xfc, 0xd8, 0x8a)


# Media types categories
MF_MT_MAJOR_TYPE = com.GUID(0x48eba18e, 0xf8c9, 0x4687, 0xbf, 0x11, 0x0a, 0x74, 0xc9, 0xf9, 0x6a, 0x8f)
MF_MT_SUBTYPE = com.GUID(0xf7e34c9a, 0x42e8, 0x4714, 0xb7, 0x4b, 0xcb, 0x29, 0xd7, 0x2c, 0x35, 0xe5)

# Major types
MFMediaType_Audio = com.GUID(0x73647561, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)
MFMediaType_Video = com.GUID(0x73646976, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xAA, 0x00, 0x38, 0x9B, 0x71)
MFMediaType_Protected = com.GUID(0x7b4b6fe6, 0x9d04, 0x4494, 0xbe, 0x14, 0x7e, 0x0b, 0xd0, 0x76, 0xc8, 0xe4)
MFMediaType_Image = com.GUID(0x72178C23, 0xE45B, 0x11D5, 0xBC, 0x2A, 0x00, 0xB0, 0xD0, 0xF3, 0xF4, 0xAB)
MFMediaType_HTML = com.GUID(0x72178C24, 0xE45B, 0x11D5, 0xBC, 0x2A, 0x00, 0xB0, 0xD0, 0xF3, 0xF4, 0xAB)
MFMediaType_Subtitle = com.GUID(0xa6d13581, 0xed50, 0x4e65, 0xae, 0x08, 0x26, 0x06, 0x55, 0x76, 0xaa, 0xcc)

# Video subtypes, attributes, and enums (Uncompressed)
D3DFMT_X8R8G8B8 = 22
D3DFMT_P8 = 41
D3DFMT_A8R8G8B8 = 21
MFVideoFormat_RGB32 = com.GUID(D3DFMT_X8R8G8B8, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)
MFVideoFormat_RGB8 = com.GUID(D3DFMT_P8, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)
MFVideoFormat_ARGB32 = com.GUID(D3DFMT_A8R8G8B8, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)

MFVideoInterlace_Progressive = 2
MF_MT_INTERLACE_MODE = com.GUID(0xe2724bb8, 0xe676, 0x4806, 0xb4, 0xb2, 0xa8, 0xd6, 0xef, 0xb4, 0x4c, 0xcd)
MF_MT_FRAME_SIZE = com.GUID(0x1652c33d, 0xd6b2, 0x4012, 0xb8, 0x34, 0x72, 0x03, 0x08, 0x49, 0xa3, 0x7d)
MF_MT_FRAME_RATE = com.GUID(0xc459a2e8, 0x3d2c, 0x4e44, 0xb1, 0x32, 0xfe, 0xe5, 0x15, 0x6c, 0x7b, 0xb0)
MF_MT_PIXEL_ASPECT_RATIO = com.GUID(0xc6376a1e, 0x8d0a, 0x4027, 0xbe, 0x45, 0x6d, 0x9a, 0x0a, 0xd3, 0x9b, 0xb6)
MF_MT_DRM_FLAGS = com.GUID(0x8772f323, 0x355a, 0x4cc7, 0xbb, 0x78, 0x6d, 0x61, 0xa0, 0x48, 0xae, 0x82)
MF_MT_DEFAULT_STRIDE = com.GUID(0x644b4e48, 0x1e02, 0x4516, 0xb0, 0xeb, 0xc0, 0x1c, 0xa9, 0xd4, 0x9a, 0xc6)

# Audio Subtypes (Uncompressed)
WAVE_FORMAT_PCM = 1
WAVE_FORMAT_IEEE_FLOAT = 3
MFAudioFormat_PCM = com.GUID(WAVE_FORMAT_PCM, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)
MFAudioFormat_Float = com.GUID(WAVE_FORMAT_IEEE_FLOAT, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)

# Image subtypes.
MFImageFormat_RGB32 = com.GUID(0x00000016, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71)
MFImageFormat_JPEG = com.GUID(0x19e4a5aa, 0x5662, 0x4fc5, 0xa0, 0xc0, 0x17, 0x58, 0x02, 0x8e, 0x10, 0x57)

# Video attributes
# Enables hardware decoding
MF_READWRITE_ENABLE_HARDWARE_TRANSFORMS = com.GUID(0xa634a91c, 0x822b, 0x41b9, 0xa4, 0x94, 0x4d, 0xe4, 0x64, 0x36, 0x12,
                                                   0xb0)
# Enable video decoding
MF_SOURCE_READER_ENABLE_VIDEO_PROCESSING = com.GUID(0xfb394f3d, 0xccf1, 0x42ee, 0xbb, 0xb3, 0xf9, 0xb8, 0x45, 0xd5,
                                                    0x68, 0x1d)
MF_SOURCE_READER_D3D_MANAGER = com.GUID(0xec822da2, 0xe1e9, 0x4b29, 0xa0, 0xd8, 0x56, 0x3c, 0x71, 0x9f, 0x52, 0x69)
MF_MEDIA_ENGINE_DXGI_MANAGER = com.GUID(0x065702da, 0x1094, 0x486d, 0x86, 0x17, 0xee, 0x7c, 0xc4, 0xee, 0x46, 0x48)
MF_SOURCE_READER_ENABLE_ADVANCED_VIDEO_PROCESSING = com.GUID(0xf81da2c, 0xb537, 0x4672, 0xa8, 0xb2, 0xa6, 0x81, 0xb1,
                                                             0x73, 0x7, 0xa3)

# Some common errors
MF_E_INVALIDSTREAMNUMBER = -1072875853  # 0xC00D36B3
MF_E_UNSUPPORTED_BYTESTREAM_TYPE = -1072875836  # 0xC00D36C4
MF_E_NO_MORE_TYPES = 0xC00D36B9
MF_E_TOPO_CODEC_NOT_FOUND = -1072868846  # 0xC00D5212


VT_I8 = 20  # Only enum we care about: https://docs.microsoft.com/en-us/windows/win32/api/wtypes/ne-wtypes-varenum


def timestamp_from_wmf(timestamp):  # 100-nanoseconds
    return float(timestamp) / 10000000


def timestamp_to_wmf(timestamp):  # 100-nanoseconds
    return int(timestamp * 10000000)


class IMFAttributes(com.pIUnknown):
    _methods_ = [
        ('GetItem',
         com.STDMETHOD()),
        ('GetItemType',
         com.STDMETHOD()),
        ('CompareItem',
         com.STDMETHOD()),
        ('Compare',
         com.STDMETHOD()),
        ('GetUINT32',
         com.STDMETHOD(com.REFIID, POINTER(c_uint32))),
        ('GetUINT64',
         com.STDMETHOD(com.REFIID, POINTER(c_uint64))),
        ('GetDouble',
         com.STDMETHOD()),
        ('GetGUID',
         com.STDMETHOD(com.REFIID, POINTER(com.GUID))),
        ('GetStringLength',
         com.STDMETHOD()),
        ('GetString',
         com.STDMETHOD()),
        ('GetAllocatedString',
         com.STDMETHOD()),
        ('GetBlobSize',
         com.STDMETHOD()),
        ('GetBlob',
         com.STDMETHOD()),
        ('GetAllocatedBlob',
         com.STDMETHOD()),
        ('GetUnknown',
         com.STDMETHOD()),
        ('SetItem',
         com.STDMETHOD()),
        ('DeleteItem',
         com.STDMETHOD()),
        ('DeleteAllItems',
         com.STDMETHOD()),
        ('SetUINT32',
         com.STDMETHOD(com.REFIID, c_uint32)),
        ('SetUINT64',
         com.STDMETHOD()),
        ('SetDouble',
         com.STDMETHOD()),
        ('SetGUID',
         com.STDMETHOD(com.REFIID, com.REFIID)),
        ('SetString',
         com.STDMETHOD()),
        ('SetBlob',
         com.STDMETHOD()),
        ('SetUnknown',
         com.STDMETHOD(com.REFIID, com.pIUnknown)),
        ('LockStore',
         com.STDMETHOD()),
        ('UnlockStore',
         com.STDMETHOD()),
        ('GetCount',
         com.STDMETHOD()),
        ('GetItemByIndex',
         com.STDMETHOD()),
        ('CopyAllItems',
         com.STDMETHOD(c_void_p)),  # IMFAttributes
    ]


class IMFMediaBuffer(com.pIUnknown):
    _methods_ = [
        ('Lock',
         com.STDMETHOD(POINTER(POINTER(BYTE)), POINTER(DWORD), POINTER(DWORD))),
        ('Unlock',
         com.STDMETHOD()),
        ('GetCurrentLength',
         com.STDMETHOD(POINTER(DWORD))),
        ('SetCurrentLength',
         com.STDMETHOD(DWORD)),
        ('GetMaxLength',
         com.STDMETHOD(POINTER(DWORD)))
    ]


class IMFSample(IMFAttributes, com.pIUnknown):
    _methods_ = [
        ('GetSampleFlags',
         com.STDMETHOD()),
        ('SetSampleFlags',
         com.STDMETHOD()),
        ('GetSampleTime',
         com.STDMETHOD()),
        ('SetSampleTime',
         com.STDMETHOD()),
        ('GetSampleDuration',
         com.STDMETHOD(POINTER(c_ulonglong))),
        ('SetSampleDuration',
         com.STDMETHOD(DWORD, IMFMediaBuffer)),
        ('GetBufferCount',
         com.STDMETHOD(POINTER(DWORD))),
        ('GetBufferByIndex',
         com.STDMETHOD(DWORD, IMFMediaBuffer)),
        ('ConvertToContiguousBuffer',
         com.STDMETHOD(POINTER(IMFMediaBuffer))),  # out
        ('AddBuffer',
         com.STDMETHOD(POINTER(DWORD))),
        ('RemoveBufferByIndex',
         com.STDMETHOD()),
        ('RemoveAllBuffers',
         com.STDMETHOD()),
        ('GetTotalLength',
         com.STDMETHOD(POINTER(DWORD))),
        ('CopyToBuffer',
         com.STDMETHOD()),
    ]


class IMFMediaType(IMFAttributes, com.pIUnknown):
    _methods_ = [
        ('GetMajorType',
         com.STDMETHOD()),
        ('IsCompressedFormat',
         com.STDMETHOD()),
        ('IsEqual',
         com.STDMETHOD()),
        ('GetRepresentation',
         com.STDMETHOD()),
        ('FreeRepresentation',
         com.STDMETHOD()),
    ]


class IMFByteStream(com.pIUnknown):
    _methods_ = [
        ('GetCapabilities',
         com.STDMETHOD()),
        ('GetLength',
         com.STDMETHOD()),
        ('SetLength',
         com.STDMETHOD()),
        ('GetCurrentPosition',
         com.STDMETHOD()),
        ('SetCurrentPosition',
         com.STDMETHOD(c_ulonglong)),
        ('IsEndOfStream',
         com.STDMETHOD()),
        ('Read',
         com.STDMETHOD()),
        ('BeginRead',
         com.STDMETHOD()),
        ('EndRead',
         com.STDMETHOD()),
        ('Write',
         com.STDMETHOD(POINTER(BYTE), ULONG, POINTER(ULONG))),
        ('BeginWrite',
         com.STDMETHOD()),
        ('EndWrite',
         com.STDMETHOD()),
        ('Seek',
         com.STDMETHOD()),
        ('Flush',
         com.STDMETHOD()),
        ('Close',
         com.STDMETHOD()),
    ]


class IMFSourceReader(com.pIUnknown):
    _methods_ = [
        ('GetStreamSelection',
         com.STDMETHOD(DWORD, POINTER(BOOL))),  # in, out
        ('SetStreamSelection',
         com.STDMETHOD(DWORD, BOOL)),
        ('GetNativeMediaType',
         com.STDMETHOD(DWORD, DWORD, POINTER(IMFMediaType))),
        ('GetCurrentMediaType',
         com.STDMETHOD(DWORD, POINTER(IMFMediaType))),
        ('SetCurrentMediaType',
         com.STDMETHOD(DWORD, POINTER(DWORD), IMFMediaType)),
        ('SetCurrentPosition',
         com.STDMETHOD(com.REFIID, POINTER(PROPVARIANT))),
        ('ReadSample',
         com.STDMETHOD(DWORD, DWORD, POINTER(DWORD), POINTER(DWORD), POINTER(c_longlong), POINTER(IMFSample))),
        ('Flush',
         com.STDMETHOD(DWORD)),  # in
        ('GetServiceForStream',
         com.STDMETHOD()),
        ('GetPresentationAttribute',
         com.STDMETHOD(DWORD, com.REFIID, POINTER(PROPVARIANT))),
    ]


class WAVEFORMATEX(ctypes.Structure):
    _fields_ = [
        ('wFormatTag', WORD),
        ('nChannels', WORD),
        ('nSamplesPerSec', DWORD),
        ('nAvgBytesPerSec', DWORD),
        ('nBlockAlign', WORD),
        ('wBitsPerSample', WORD),
        ('cbSize', WORD),
    ]

    def __repr__(self):
        return 'WAVEFORMATEX(wFormatTag={}, nChannels={}, nSamplesPerSec={}, nAvgBytesPersec={}' \
               ', nBlockAlign={}, wBitsPerSample={}, cbSize={})'.format(
            self.wFormatTag, self.nChannels, self.nSamplesPerSec,
            self.nAvgBytesPerSec, self.nBlockAlign, self.wBitsPerSample,
            self.cbSize)


# Stream constants
MF_SOURCE_READER_ALL_STREAMS = 0xfffffffe
MF_SOURCE_READER_ANY_STREAM = 4294967294  # 0xfffffffe
MF_SOURCE_READER_FIRST_AUDIO_STREAM = 4294967293  # 0xfffffffd
MF_SOURCE_READER_FIRST_VIDEO_STREAM = 0xfffffffc
MF_SOURCE_READER_MEDIASOURCE = 0xffffffff

# Version calculation
if WINDOWS_7_OR_GREATER:
    MF_SDK_VERSION = 0x0002
else:
    MF_SDK_VERSION = 0x0001

MF_API_VERSION = 0x0070  # Only used in Vista.

MF_VERSION = (MF_SDK_VERSION << 16 | MF_API_VERSION)

MFStartup = mfplat_lib.MFStartup
MFStartup.restype = HRESULT
MFStartup.argtypes = [LONG, DWORD]

MFShutdown = mfplat_lib.MFShutdown
MFShutdown.restype = HRESULT
MFShutdown.argtypes = []

MFCreateAttributes = mfplat_lib.MFCreateAttributes
MFCreateAttributes.restype = HRESULT
MFCreateAttributes.argtypes = [POINTER(IMFAttributes), c_uint32]  # attributes, cInitialSize

MFCreateSourceReaderFromURL = mfreadwrite_lib.MFCreateSourceReaderFromURL
MFCreateSourceReaderFromURL.restype = HRESULT
MFCreateSourceReaderFromURL.argtypes = [LPCWSTR, IMFAttributes, POINTER(IMFSourceReader)]

MFCreateSourceReaderFromByteStream = mfreadwrite_lib.MFCreateSourceReaderFromByteStream
MFCreateSourceReaderFromByteStream.restype = HRESULT
MFCreateSourceReaderFromByteStream.argtypes = [IMFByteStream, IMFAttributes, POINTER(IMFSourceReader)]

if WINDOWS_7_OR_GREATER:
    MFCreateMFByteStreamOnStream = mfplat_lib.MFCreateMFByteStreamOnStream
    MFCreateMFByteStreamOnStream.restype = HRESULT
    MFCreateMFByteStreamOnStream.argtypes = [c_void_p, POINTER(IMFByteStream)]

MFCreateTempFile = mfplat_lib.MFCreateTempFile
MFCreateTempFile.restype = HRESULT
MFCreateTempFile.argtypes = [UINT, UINT, UINT, POINTER(IMFByteStream)]

MFCreateMediaType = mfplat_lib.MFCreateMediaType
MFCreateMediaType.restype = HRESULT
MFCreateMediaType.argtypes = [POINTER(IMFMediaType)]

MFCreateWaveFormatExFromMFMediaType = mfplat_lib.MFCreateWaveFormatExFromMFMediaType
MFCreateWaveFormatExFromMFMediaType.restype = HRESULT
MFCreateWaveFormatExFromMFMediaType.argtypes = [IMFMediaType, POINTER(POINTER(WAVEFORMATEX)), POINTER(c_uint32), c_uint32]


class WMFSource(Source):
    low_latency = True  # Quicker latency but possible quality loss.

    decode_audio = True
    decode_video = True

    def __init__(self, filename, file=None):
        assert any([self.decode_audio, self.decode_video]), "Source must decode audio, video, or both, not none."
        self._current_audio_sample = None
        self._current_audio_buffer = None
        self._current_video_sample = None
        self._current_video_buffer = None
        self._timestamp = 0
        self._attributes = None
        self._stream_obj = None
        self._imf_bytestream = None
        self._wfx = None
        self._stride = None

        self.set_config_attributes()

        # Create SourceReader
        self._source_reader = IMFSourceReader()

        # If it's a file, we need to load it as a stream.
        if file is not None:
            data = file.read()

            self._imf_bytestream = IMFByteStream()

            data_len = len(data)

            if WINDOWS_7_OR_GREATER:
                # Stole code from GDIPlus for older IStream support.
                hglob = kernel32.GlobalAlloc(GMEM_MOVEABLE, data_len)
                ptr = kernel32.GlobalLock(hglob)
                ctypes.memmove(ptr, data, data_len)
                kernel32.GlobalUnlock(hglob)

                # Create IStream
                self._stream_obj = com.pIUnknown()
                ole32.CreateStreamOnHGlobal(hglob, True, ctypes.byref(self._stream_obj))

                # MFCreateMFByteStreamOnStreamEx for future async operations exists, however Windows 8+ only. Requires new interface
                # (Also unsure how/if new Windows async functions and callbacks work with ctypes.)
                MFCreateMFByteStreamOnStream(self._stream_obj, ctypes.byref(self._imf_bytestream))
            else:
                # Vista does not support MFCreateMFByteStreamOnStream.
                # HACK: Create file in Windows temp folder to write our byte data to.
                # (Will be automatically deleted when IMFByteStream is Released.)
                MFCreateTempFile(MF_ACCESSMODE_READWRITE,
                                 MF_OPENMODE_DELETE_IF_EXIST,
                                 MF_FILEFLAGS_NONE,
                                 ctypes.byref(self._imf_bytestream))

                wrote_length = ULONG()
                data_ptr = cast(data, POINTER(BYTE))
                self._imf_bytestream.Write(data_ptr, data_len, ctypes.byref(wrote_length))
                self._imf_bytestream.SetCurrentPosition(0)

                if wrote_length.value != data_len:
                    raise MediaDecodeException("Could not write all of the data to the bytestream file.")

            try:
                MFCreateSourceReaderFromByteStream(self._imf_bytestream, self._attributes, ctypes.byref(self._source_reader))
            except OSError as err:
                raise MediaDecodeException(err) from None
        else:
            # We can just load from filename if no file object specified..
            try:
                MFCreateSourceReaderFromURL(filename, self._attributes, ctypes.byref(self._source_reader))
            except OSError as err:
                raise MediaDecodeException(err) from None

        if self.decode_audio:
            self._load_audio()

        if self.decode_video:
            self._load_video()

        assert self.audio_format or self.video_format, "Source was decoded, but no video or audio streams were found."

        # Get duration of the media file after everything has been ok to decode.
        try:
            prop = PROPVARIANT()
            self._source_reader.GetPresentationAttribute(MF_SOURCE_READER_MEDIASOURCE,
                                                         ctypes.byref(MF_PD_DURATION),
                                                         ctypes.byref(prop))

            self._duration = timestamp_from_wmf(prop.llVal)
            ole32.PropVariantClear(ctypes.byref(prop))
        except OSError:
            warnings.warn("Could not determine duration of media file: '{}'.".format(filename))

    def _load_audio(self, stream=MF_SOURCE_READER_FIRST_AUDIO_STREAM):
        """ Prepares the audio stream for playback by detecting if it's compressed and attempting to decompress to PCM.
            Default: Only get the first available audio stream.
        """
        # Will be an audio file.
        self._audio_stream_index = stream

        # Get what the native/real media type is (audio only)
        imfmedia = IMFMediaType()

        try:
            self._source_reader.GetNativeMediaType(self._audio_stream_index, 0, ctypes.byref(imfmedia))
        except OSError as err:
            if err.winerror == MF_E_INVALIDSTREAMNUMBER:
                assert _debug('WMFAudioDecoder: No audio stream found.')
            return

        # Get Major media type (Audio, Video, etc)
        # TODO: Make GUID take no arguments for a null version:
        guid_audio_type = com.GUID(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

        imfmedia.GetGUID(MF_MT_MAJOR_TYPE, ctypes.byref(guid_audio_type))

        if guid_audio_type == MFMediaType_Audio:
            assert _debug('WMFAudioDecoder: Found Audio Stream.')

            # Deselect any other streams if we don't need them. (Small speedup)
            if not self.decode_video:
                self._source_reader.SetStreamSelection(MF_SOURCE_READER_ANY_STREAM, False)

            # Select first audio stream.
            self._source_reader.SetStreamSelection(MF_SOURCE_READER_FIRST_AUDIO_STREAM, True)

            # Check sub media type, AKA what kind of codec
            guid_compressed = com.GUID(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
            imfmedia.GetGUID(MF_MT_SUBTYPE, ctypes.byref(guid_compressed))

            if guid_compressed == MFAudioFormat_PCM or guid_compressed == MFAudioFormat_Float:
                assert _debug('WMFAudioDecoder: Found Uncompressed Audio:', guid_compressed)
            else:
                assert _debug('WMFAudioDecoder: Found Compressed Audio:', guid_compressed)
                # If audio is compressed, attempt to decompress it by forcing source reader to use PCM
                mf_mediatype = IMFMediaType()

                MFCreateMediaType(ctypes.byref(mf_mediatype))
                mf_mediatype.SetGUID(MF_MT_MAJOR_TYPE, MFMediaType_Audio)
                mf_mediatype.SetGUID(MF_MT_SUBTYPE, MFAudioFormat_PCM)

                try:
                    self._source_reader.SetCurrentMediaType(self._audio_stream_index, None, mf_mediatype)
                except OSError as err:  # Can't decode codec.
                    raise MediaDecodeException(err) from None

            # Current media type should now be properly decoded at this point.
            decoded_media_type = IMFMediaType()  # Maybe reusing older IMFMediaType will work?
            self._source_reader.GetCurrentMediaType(self._audio_stream_index, ctypes.byref(decoded_media_type))

            wfx_length = ctypes.c_uint32()
            wfx = POINTER(WAVEFORMATEX)()

            MFCreateWaveFormatExFromMFMediaType(decoded_media_type,
                                                ctypes.byref(wfx),
                                                ctypes.byref(wfx_length),
                                                0)

            self._wfx = wfx.contents
            self.audio_format = AudioFormat(channels=self._wfx.nChannels,
                                            sample_size=self._wfx.wBitsPerSample,
                                            sample_rate=self._wfx.nSamplesPerSec)
        else:
            assert _debug('WMFAudioDecoder: Audio stream not found')

    def get_format(self):
        """Returns the WAVEFORMATEX data which has more information thah audio_format"""
        return self._wfx

    def _load_video(self, stream=MF_SOURCE_READER_FIRST_VIDEO_STREAM):
        self._video_stream_index = stream

        # Get what the native/real media type is (video only)
        imfmedia = IMFMediaType()

        try:
            self._source_reader.GetCurrentMediaType(self._video_stream_index, ctypes.byref(imfmedia))
        except OSError as err:
            if err.winerror == MF_E_INVALIDSTREAMNUMBER:
                assert _debug('WMFVideoDecoder: No video stream found.')
            return

        assert _debug('WMFVideoDecoder: Found Video Stream')

        # All video is basically compressed, try to decompress.
        uncompressed_mt = IMFMediaType()
        MFCreateMediaType(ctypes.byref(uncompressed_mt))

        imfmedia.CopyAllItems(uncompressed_mt)

        imfmedia.Release()

        uncompressed_mt.SetGUID(MF_MT_SUBTYPE, MFVideoFormat_RGB32)
        uncompressed_mt.SetUINT32(MF_MT_INTERLACE_MODE, MFVideoInterlace_Progressive)
        uncompressed_mt.SetUINT32(MF_MT_ALL_SAMPLES_INDEPENDENT, 1)

        try:
            self._source_reader.SetCurrentMediaType(self._video_stream_index, None, uncompressed_mt)
        except OSError as err:  # Can't decode codec.
            raise MediaDecodeException(err) from None

        height, width = self._get_attribute_size(uncompressed_mt, MF_MT_FRAME_SIZE)

        self.video_format = VideoFormat(width=width, height=height)
        assert _debug('WMFVideoDecoder: Frame width: {} height: {}'.format(width, height))

        # Frame rate
        den, num = self._get_attribute_size(uncompressed_mt, MF_MT_FRAME_RATE)
        self.video_format.frame_rate = num / den
        assert _debug('WMFVideoDecoder: Frame Rate: {} / {} = {}'.format(num, den, self.video_format.frame_rate))

        # Sometimes it can return negative? Variable bit rate? Needs further tests and examples.
        if self.video_format.frame_rate < 0:
            self.video_format.frame_rate = 30000 / 1001
            assert _debug('WARNING: Negative frame rate, attempting to use default, but may experience issues.')

        # Pixel ratio
        den, num = self._get_attribute_size(uncompressed_mt, MF_MT_PIXEL_ASPECT_RATIO)
        self.video_format.sample_aspect = num / den
        assert _debug('WMFVideoDecoder: Pixel Ratio: {} / {} = {}'.format(num, den, self.video_format.sample_aspect))

    def get_audio_data(self, num_bytes, compensation_time=0.0):
        flags = DWORD()
        timestamp = ctypes.c_longlong()
        audio_data_length = DWORD()

        # If we have an audio sample already in use and we call this again, release the memory of buffer and sample.
        # Can only release after the data is played or else glitches and pops can be heard.
        if self._current_audio_sample:
            self._current_audio_buffer.Release()
            self._current_audio_sample.Release()

        self._current_audio_sample = IMFSample()
        self._current_audio_buffer = IMFMediaBuffer()

        while True:
            self._source_reader.ReadSample(self._audio_stream_index, 0, None, ctypes.byref(flags),
                                           ctypes.byref(timestamp), ctypes.byref(self._current_audio_sample))

            if flags.value & MF_SOURCE_READERF_CURRENTMEDIATYPECHANGED:
                assert _debug('WMFAudioDecoder: Data is no longer valid.')
                break

            if flags.value & MF_SOURCE_READERF_ENDOFSTREAM:
                assert _debug('WMFAudioDecoder: End of data from stream source.')
                break

            if not self._current_audio_sample:
                assert _debug('WMFAudioDecoder: No sample.')
                continue

            # Convert to single buffer as a sample could potentially(rarely) have multiple buffers.
            self._current_audio_sample.ConvertToContiguousBuffer(ctypes.byref(self._current_audio_buffer))

            audio_data_ptr = POINTER(BYTE)()

            self._current_audio_buffer.Lock(ctypes.byref(audio_data_ptr), None, ctypes.byref(audio_data_length))
            self._current_audio_buffer.Unlock()

            audio_data = create_string_buffer(audio_data_length.value)
            memmove(audio_data, audio_data_ptr, audio_data_length.value)

            return AudioData(audio_data,
                             audio_data_length.value,
                             timestamp_from_wmf(timestamp.value),
                             audio_data_length.value / self.audio_format.sample_rate,
                             [])

        return None

    def get_next_video_frame(self, skip_empty_frame=True):
        video_data_length = DWORD()
        flags = DWORD()
        timestamp = ctypes.c_longlong()

        if self._current_video_sample:
            self._current_video_buffer.Release()
            self._current_video_sample.Release()

        self._current_video_sample = IMFSample()
        self._current_video_buffer = IMFMediaBuffer()

        while True:
            self._source_reader.ReadSample(self._video_stream_index, 0, None, ctypes.byref(flags),
                                           ctypes.byref(timestamp), ctypes.byref(self._current_video_sample))

            if flags.value & MF_SOURCE_READERF_CURRENTMEDIATYPECHANGED:
                assert _debug('WMFVideoDecoder: Data is no longer valid.')

                # Get Major media type (Audio, Video, etc)
                new = IMFMediaType()
                self._source_reader.GetCurrentMediaType(self._video_stream_index, ctypes.byref(new))

                # Sometimes this happens once. I think this only
                # changes if the stride is added/changed before playback?
                stride = ctypes.c_uint32()
                new.GetUINT32(MF_MT_DEFAULT_STRIDE, ctypes.byref(stride))

                self._stride = stride.value

            if flags.value & MF_SOURCE_READERF_ENDOFSTREAM:
                self._timestamp = None
                assert _debug('WMFVideoDecoder: End of data from stream source.')
                break

            if not self._current_video_sample:
                assert _debug('WMFVideoDecoder: No sample.')
                continue

            self._current_video_buffer = IMFMediaBuffer()

            # Convert to single buffer as a sample could potentially have multiple buffers.
            self._current_video_sample.ConvertToContiguousBuffer(ctypes.byref(self._current_video_buffer))

            video_data = POINTER(BYTE)()

            self._current_video_buffer.Lock(ctypes.byref(video_data), None, ctypes.byref(video_data_length))

            width = self.video_format.width
            height = self.video_format.height

            # buffer = ctypes.create_string_buffer(size)
            self._timestamp = timestamp_from_wmf(timestamp.value)

            self._current_video_buffer.Unlock()

            # This is made with the assumption that the video frame will be blitted into the player texture immediately
            # after, and then cleared next frame attempt.
            return image.ImageData(width, height, 'BGRA', video_data, self._stride)

        return None

    def get_next_video_timestamp(self):
        return self._timestamp

    def seek(self, timestamp):
        timestamp = min(timestamp, self._duration) if self._duration else timestamp

        prop = PROPVARIANT()
        prop.vt = VT_I8
        prop.llVal = timestamp_to_wmf(timestamp)

        pos_com = com.GUID(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        try:
            self._source_reader.SetCurrentPosition(pos_com, prop)
        except OSError as err:
            warnings.warn(str(err))

        ole32.PropVariantClear(ctypes.byref(prop))

    @staticmethod
    def _get_attribute_size(attributes, guidKey):
        """ Convert int64 attributes to int32"""  # HI32/LOW32

        size = ctypes.c_uint64()
        attributes.GetUINT64(guidKey, size)
        lParam = size.value

        x = ctypes.c_int32(lParam).value
        y = ctypes.c_int32(lParam >> 32).value
        return x, y

    def set_config_attributes(self):
        """ Here we set user specified attributes, by default we try to set low latency mode. (Win7+)"""
        if self.low_latency or self.decode_video:
            self._attributes = IMFAttributes()

            MFCreateAttributes(ctypes.byref(self._attributes), 3)

        if self.low_latency and WINDOWS_7_OR_GREATER:
            self._attributes.SetUINT32(ctypes.byref(MF_LOW_LATENCY), 1)

            assert _debug('WMFAudioDecoder: Setting configuration attributes.')

        # If it's a video we need to enable the streams to be accessed.
        if self.decode_video:
            self._attributes.SetUINT32(ctypes.byref(MF_READWRITE_ENABLE_HARDWARE_TRANSFORMS), 1)
            self._attributes.SetUINT32(ctypes.byref(MF_SOURCE_READER_ENABLE_VIDEO_PROCESSING), 1)

            assert _debug('WMFVideoDecoder: Setting configuration attributes.')

    def __del__(self):
        if self._stream_obj:
            self._stream_obj.Release()

        if self._imf_bytestream:
            self._imf_bytestream.Release()

        if self._current_audio_sample:
            self._current_audio_buffer.Release()
            self._current_audio_sample.Release()

        if self._current_video_sample:
            self._current_video_buffer.Release()
            self._current_video_sample.Release()


#########################################
#   Decoder class:
#########################################

class WMFDecoder(MediaDecoder):
    def __init__(self):

        self.ole32 = None
        self.MFShutdown = None

        try:
            # Coinitialize supposed to be called for COMs?
            ole32.CoInitializeEx(None, COINIT_MULTITHREADED)
        except OSError as err:
            warnings.warn(str(err))

        try:
            MFStartup(MF_VERSION, 0)
        except OSError as err:
            raise ImportError('WMF could not startup:', err.strerror)

        self.extensions = self._build_decoder_extensions()

        self.ole32 = ole32
        self.MFShutdown = MFShutdown

        assert _debug('Windows Media Foundation: Initialized.')

    @staticmethod
    def _build_decoder_extensions():
        """Extension support varies depending on OS version."""
        extensions = []
        if WINDOWS_VISTA_OR_GREATER:
            extensions.extend(['.asf', '.wma', '.wmv',
                               '.mp3',
                               '.sami', '.smi',
                               ])

        if WINDOWS_7_OR_GREATER:
            extensions.extend(['.3g2', '.3gp', '.3gp2', '.3gp',
                               '.aac', '.adts',
                               '.avi',
                               '.m4a', '.m4v', '.mov', '.mp4',
                               # '.wav'  # Can do wav, but we have a WAVE decoder.
                               ])

        if WINDOWS_10_ANNIVERSARY_UPDATE_OR_GREATER:
            extensions.extend(['.mkv', '.flac', '.ogg'])

        return extensions

    def get_file_extensions(self):
        return self.extensions

    def decode(self, file, filename, streaming=True):
        if streaming:
            return WMFSource(filename, file)
        else:
            return StaticSource(WMFSource(filename, file))

    def __del__(self):
        if self.MFShutdown is not None:
            self.MFShutdown()
        if self.ole32 is not None:
            self.ole32.CoUninitialize()


def get_decoders():
    return [WMFDecoder()]


def get_encoders():
    return []



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

import os.path

from pyglet.image import *
from pyglet.image.codecs import *

try:
    import Image
except ImportError:
    from PIL import Image, ImageSequence


class PILImageDecoder(ImageDecoder):
    def get_file_extensions(self):
        # Only most common ones shown here
        return ['.bmp', '.cur', '.gif', '.ico', '.jpg', '.jpeg', '.pcx', '.png',
                '.tga', '.tif', '.tiff', '.xbm', '.xpm']

    # def get_animation_file_extensions(self):
    #     return ['.gif', '.ani']

    def decode(self, file, filename):
        try:
            image = Image.open(file)
        except Exception as e:
            raise ImageDecodeException(
                'PIL cannot read %r: %s' % (filename or file, e))

        try:
            image = image.transpose(Image.FLIP_TOP_BOTTOM)
        except Exception as e:
            raise ImageDecodeException('PIL failed to transpose %r: %s' % (filename or file, e))

        # Convert bitmap and palette images to component
        if image.mode in ('1', 'P'):
            image = image.convert()

        if image.mode not in ('L', 'LA', 'RGB', 'RGBA'):
            raise ImageDecodeException('Unsupported mode "%s"' % image.mode)
        width, height = image.size

        # tostring is deprecated, replaced by tobytes in Pillow (PIL fork)
        # (1.1.7) PIL still uses it
        try:
            image_data_fn = getattr(image, "tobytes")
        except AttributeError:
            image_data_fn = getattr(image, "tostring")
        return ImageData(width, height, image.mode, image_data_fn())

    # def decode_animation(self, file, filename):
    #     try:
    #         image = Image.open(file)
    #     except Exception as e:
    #         raise ImageDecodeException('PIL cannot read %r: %s' % (filename or file, e))
    #
    #     frames = []
    #
    #     for image in ImageSequence.Iterator(image):
    #         try:
    #             image = image.transpose(Image.FLIP_TOP_BOTTOM)
    #         except Exception as e:
    #             raise ImageDecodeException('PIL failed to transpose %r: %s' % (filename or file, e))
    #
    #         # Convert bitmap and palette images to component
    #         if image.mode in ('1', 'P'):
    #             image = image.convert()
    #
    #         if image.mode not in ('L', 'LA', 'RGB', 'RGBA'):
    #             raise ImageDecodeException('Unsupported mode "%s"' % image.mode)
    #
    #         duration = None if image.info['duration'] == 0 else image.info['duration']
    #         # Follow Firefox/Mac behaviour: use 100ms delay for any delay less than 10ms.
    #         if duration <= 10:
    #             duration = 100
    #
    #         frames.append(AnimationFrame(ImageData(*image.size, image.mode, image.tobytes()), duration / 1000))
    #
    #     return Animation(frames)


class PILImageEncoder(ImageEncoder):
    def get_file_extensions(self):
        # Most common only
        return ['.bmp', '.eps', '.gif', '.jpg', '.jpeg',
                '.pcx', '.png', '.ppm', '.tiff', '.xbm']

    def encode(self, image, file, filename):
        # File format is guessed from filename extension, otherwise defaults
        # to PNG.
        pil_format = (filename and os.path.splitext(filename)[1][1:]) or 'png'

        if pil_format.lower() == 'jpg':
            pil_format = 'JPEG'

        image = image.get_image_data()
        fmt = image.format
        if fmt != 'RGB':
            # Only save in RGB or RGBA formats.
            fmt = 'RGBA'
        pitch = -(image.width * len(fmt))

        # fromstring is deprecated, replaced by frombytes in Pillow (PIL fork)
        # (1.1.7) PIL still uses it
        try:
            image_from_fn = getattr(Image, "frombytes")
        except AttributeError:
            image_from_fn = getattr(Image, "fromstring")
        pil_image = image_from_fn(fmt, (image.width, image.height), image.get_data(fmt, pitch))

        try:
            pil_image.save(file, pil_format)
        except Exception as e:
            raise ImageEncodeException(e)


def get_decoders():
    return [PILImageDecoder()]


def get_encoders():
    return [PILImageEncoder()]



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

import pyglet
import warnings
from pyglet import app
from .base import Display, Screen, ScreenMode, Canvas


from ctypes import *
from pyglet.libs.egl import egl
from pyglet.libs.egl import eglext


class HeadlessDisplay(Display):

    def __init__(self):
        super().__init__()
        # TODO: fix this placeholder:
        self._screens = [HeadlessScreen(self, 0, 0, 1920, 1080)]

        num_devices = egl.EGLint()
        eglext.eglQueryDevicesEXT(0, None, byref(num_devices))
        if num_devices.value > 0:
            headless_device = pyglet.options['headless_device']
            if headless_device < 0 or headless_device >= num_devices.value:
                 raise ValueError('Invalid EGL devide id: %d' % headless_device)
            devices = (eglext.EGLDeviceEXT * num_devices.value)()
            eglext.eglQueryDevicesEXT(num_devices.value, devices, byref(num_devices))
            self._display_connection  = eglext.eglGetPlatformDisplayEXT(eglext.EGL_PLATFORM_DEVICE_EXT, devices[headless_device], None)
        else:
            warning.warn('No device available for EGL device platform. Using native display type.')
            display = egl.EGLNativeDisplayType()
            self._display_connection = egl.eglGetDisplay(display)

        egl.eglInitialize(self._display_connection, None, None)

    def get_screens(self):
        return self._screens

    def __del__(self):
        egl.eglTerminate(self._display_connection)


class HeadlessCanvas(Canvas):
    def __init__(self, display, egl_surface):
        super(HeadlessCanvas, self).__init__(display)
        self.egl_surface = egl_surface

class HeadlessScreen(Screen):
    def __init__(self, display, x, y, width, height):
        super().__init__(display, x, y, width, height)

    def get_matching_configs(self, template):
        canvas = HeadlessCanvas(self.display, None)
        configs = template.match(canvas)
        # XXX deprecate
        for config in configs:
            config.screen = self
        # print("Canvas", canvas, "configs", configs)
        return configs

    def get_modes(self):
        pass

    def get_mode(self):
        pass

    def set_mode(self, mode):
        pass

    def restore_mode(self):
        pass



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# Copyright (c) 2008-2020 pyglet contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
"""Wrapper for Xi

Generated with:
tools/genwrappers.py xinput

Do not modify this file.
"""

import ctypes
from ctypes import *

import pyglet.lib

_lib = pyglet.lib.load_library('Xi')

_int_types = (c_int16, c_int32)
if hasattr(ctypes, 'c_int64'):
    # Some builds of ctypes apparently do not have c_int64
    # defined; it's a pretty good bet that these builds do not
    # have 64-bit pointers.
    _int_types += (ctypes.c_int64,)
for t in _int_types:
    if sizeof(t) == sizeof(c_size_t):
        c_ptrdiff_t = t

class c_void(Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', c_int)]


import pyglet.libs.x11.xlib

sz_xGetExtensionVersionReq = 8 	# /usr/include/X11/extensions/XI.h:56
sz_xGetExtensionVersionReply = 32 	# /usr/include/X11/extensions/XI.h:57
sz_xListInputDevicesReq = 4 	# /usr/include/X11/extensions/XI.h:58
sz_xListInputDevicesReply = 32 	# /usr/include/X11/extensions/XI.h:59
sz_xListDevicePropertiesReq = 8     # /usr/include/X11/extensions/XI.h
sz_xListDevicePropertiesReply = 32  # /usr/include/X11/extensions/XI.h
sz_xGetDevicePropertyReq = 24     # /usr/include/X11/extensions/XI.h
sz_xGetDevicePropertyReply = 32  # /usr/include/X11/extensions/XI.h
sz_xOpenDeviceReq = 8 	# /usr/include/X11/extensions/XI.h:60
sz_xOpenDeviceReply = 32 	# /usr/include/X11/extensions/XI.h:61
sz_xCloseDeviceReq = 8 	# /usr/include/X11/extensions/XI.h:62
sz_xSetDeviceModeReq = 8 	# /usr/include/X11/extensions/XI.h:63
sz_xSetDeviceModeReply = 32 	# /usr/include/X11/extensions/XI.h:64
sz_xSelectExtensionEventReq = 12 	# /usr/include/X11/extensions/XI.h:65
sz_xGetSelectedExtensionEventsReq = 8 	# /usr/include/X11/extensions/XI.h:66
sz_xGetSelectedExtensionEventsReply = 32 	# /usr/include/X11/extensions/XI.h:67
sz_xChangeDeviceDontPropagateListReq = 12 	# /usr/include/X11/extensions/XI.h:68
sz_xGetDeviceDontPropagateListReq = 8 	# /usr/include/X11/extensions/XI.h:69
sz_xGetDeviceDontPropagateListReply = 32 	# /usr/include/X11/extensions/XI.h:70
sz_xGetDeviceMotionEventsReq = 16 	# /usr/include/X11/extensions/XI.h:71
sz_xGetDeviceMotionEventsReply = 32 	# /usr/include/X11/extensions/XI.h:72
sz_xChangeKeyboardDeviceReq = 8 	# /usr/include/X11/extensions/XI.h:73
sz_xChangeKeyboardDeviceReply = 32 	# /usr/include/X11/extensions/XI.h:74
sz_xChangePointerDeviceReq = 8 	# /usr/include/X11/extensions/XI.h:75
sz_xChangePointerDeviceReply = 32 	# /usr/include/X11/extensions/XI.h:76
sz_xGrabDeviceReq = 20 	# /usr/include/X11/extensions/XI.h:77
sz_xGrabDeviceReply = 32 	# /usr/include/X11/extensions/XI.h:78
sz_xUngrabDeviceReq = 12 	# /usr/include/X11/extensions/XI.h:79
sz_xGrabDeviceKeyReq = 20 	# /usr/include/X11/extensions/XI.h:80
sz_xGrabDeviceKeyReply = 32 	# /usr/include/X11/extensions/XI.h:81
sz_xUngrabDeviceKeyReq = 16 	# /usr/include/X11/extensions/XI.h:82
sz_xGrabDeviceButtonReq = 20 	# /usr/include/X11/extensions/XI.h:83
sz_xGrabDeviceButtonReply = 32 	# /usr/include/X11/extensions/XI.h:84
sz_xUngrabDeviceButtonReq = 16 	# /usr/include/X11/extensions/XI.h:85
sz_xAllowDeviceEventsReq = 12 	# /usr/include/X11/extensions/XI.h:86
sz_xGetDeviceFocusReq = 8 	# /usr/include/X11/extensions/XI.h:87
sz_xGetDeviceFocusReply = 32 	# /usr/include/X11/extensions/XI.h:88
sz_xSetDeviceFocusReq = 16 	# /usr/include/X11/extensions/XI.h:89
sz_xGetFeedbackControlReq = 8 	# /usr/include/X11/extensions/XI.h:90
sz_xGetFeedbackControlReply = 32 	# /usr/include/X11/extensions/XI.h:91
sz_xChangeFeedbackControlReq = 12 	# /usr/include/X11/extensions/XI.h:92
sz_xGetDeviceKeyMappingReq = 8 	# /usr/include/X11/extensions/XI.h:93
sz_xGetDeviceKeyMappingReply = 32 	# /usr/include/X11/extensions/XI.h:94
sz_xChangeDeviceKeyMappingReq = 8 	# /usr/include/X11/extensions/XI.h:95
sz_xGetDeviceModifierMappingReq = 8 	# /usr/include/X11/extensions/XI.h:96
sz_xSetDeviceModifierMappingReq = 8 	# /usr/include/X11/extensions/XI.h:97
sz_xSetDeviceModifierMappingReply = 32 	# /usr/include/X11/extensions/XI.h:98
sz_xGetDeviceButtonMappingReq = 8 	# /usr/include/X11/extensions/XI.h:99
sz_xGetDeviceButtonMappingReply = 32 	# /usr/include/X11/extensions/XI.h:100
sz_xSetDeviceButtonMappingReq = 8 	# /usr/include/X11/extensions/XI.h:101
sz_xSetDeviceButtonMappingReply = 32 	# /usr/include/X11/extensions/XI.h:102
sz_xQueryDeviceStateReq = 8 	# /usr/include/X11/extensions/XI.h:103
sz_xQueryDeviceStateReply = 32 	# /usr/include/X11/extensions/XI.h:104
sz_xSendExtensionEventReq = 16 	# /usr/include/X11/extensions/XI.h:105
sz_xDeviceBellReq = 8 	# /usr/include/X11/extensions/XI.h:106
sz_xSetDeviceValuatorsReq = 8 	# /usr/include/X11/extensions/XI.h:107
sz_xSetDeviceValuatorsReply = 32 	# /usr/include/X11/extensions/XI.h:108
sz_xGetDeviceControlReq = 8 	# /usr/include/X11/extensions/XI.h:109
sz_xGetDeviceControlReply = 32 	# /usr/include/X11/extensions/XI.h:110
sz_xChangeDeviceControlReq = 8 	# /usr/include/X11/extensions/XI.h:111
sz_xChangeDeviceControlReply = 32 	# /usr/include/X11/extensions/XI.h:112
Dont_Check = 0 	# /usr/include/X11/extensions/XI.h:135
XInput_Initial_Release = 1 	# /usr/include/X11/extensions/XI.h:136
XInput_Add_XDeviceBell = 2 	# /usr/include/X11/extensions/XI.h:137
XInput_Add_XSetDeviceValuators = 3 	# /usr/include/X11/extensions/XI.h:138
XInput_Add_XChangeDeviceControl = 4 	# /usr/include/X11/extensions/XI.h:139
XInput_Add_DevicePresenceNotify = 5 	# /usr/include/X11/extensions/XI.h:140
XI_Absent = 0 	# /usr/include/X11/extensions/XI.h:142
XI_Present = 1 	# /usr/include/X11/extensions/XI.h:143
XI_Initial_Release_Major = 1 	# /usr/include/X11/extensions/XI.h:145
XI_Initial_Release_Minor = 0 	# /usr/include/X11/extensions/XI.h:146
XI_Add_XDeviceBell_Major = 1 	# /usr/include/X11/extensions/XI.h:148
XI_Add_XDeviceBell_Minor = 1 	# /usr/include/X11/extensions/XI.h:149
XI_Add_XSetDeviceValuators_Major = 1 	# /usr/include/X11/extensions/XI.h:151
XI_Add_XSetDeviceValuators_Minor = 2 	# /usr/include/X11/extensions/XI.h:152
XI_Add_XChangeDeviceControl_Major = 1 	# /usr/include/X11/extensions/XI.h:154
XI_Add_XChangeDeviceControl_Minor = 3 	# /usr/include/X11/extensions/XI.h:155
XI_Add_DevicePresenceNotify_Major = 1 	# /usr/include/X11/extensions/XI.h:157
XI_Add_DevicePresenceNotify_Minor = 4 	# /usr/include/X11/extensions/XI.h:158
DEVICE_RESOLUTION = 1 	# /usr/include/X11/extensions/XI.h:160
DEVICE_ABS_CALIB = 2 	# /usr/include/X11/extensions/XI.h:161
DEVICE_CORE = 3 	# /usr/include/X11/extensions/XI.h:162
DEVICE_ENABLE = 4 	# /usr/include/X11/extensions/XI.h:163
DEVICE_ABS_AREA = 5 	# /usr/include/X11/extensions/XI.h:164
NoSuchExtension = 1 	# /usr/include/X11/extensions/XI.h:166
COUNT = 0 	# /usr/include/X11/extensions/XI.h:168
CREATE = 1 	# /usr/include/X11/extensions/XI.h:169
NewPointer = 0 	# /usr/include/X11/extensions/XI.h:171
NewKeyboard = 1 	# /usr/include/X11/extensions/XI.h:172
XPOINTER = 0 	# /usr/include/X11/extensions/XI.h:174
XKEYBOARD = 1 	# /usr/include/X11/extensions/XI.h:175
UseXKeyboard = 255 	# /usr/include/X11/extensions/XI.h:177
IsXPointer = 0 	# /usr/include/X11/extensions/XI.h:179
IsXKeyboard = 1 	# /usr/include/X11/extensions/XI.h:180
IsXExtensionDevice = 2 	# /usr/include/X11/extensions/XI.h:181
IsXExtensionKeyboard = 3 	# /usr/include/X11/extensions/XI.h:182
IsXExtensionPointer = 4 	# /usr/include/X11/extensions/XI.h:183
AsyncThisDevice = 0 	# /usr/include/X11/extensions/XI.h:185
SyncThisDevice = 1 	# /usr/include/X11/extensions/XI.h:186
ReplayThisDevice = 2 	# /usr/include/X11/extensions/XI.h:187
AsyncOtherDevices = 3 	# /usr/include/X11/extensions/XI.h:188
AsyncAll = 4 	# /usr/include/X11/extensions/XI.h:189
SyncAll = 5 	# /usr/include/X11/extensions/XI.h:190
FollowKeyboard = 3 	# /usr/include/X11/extensions/XI.h:192
RevertToFollowKeyboard = 3 	# /usr/include/X11/extensions/XI.h:194
DvAccelNum = 1 	# /usr/include/X11/extensions/XI.h:197
DvAccelDenom = 2 	# /usr/include/X11/extensions/XI.h:198
DvThreshold = 4 	# /usr/include/X11/extensions/XI.h:199
DvKeyClickPercent = 1 	# /usr/include/X11/extensions/XI.h:201
DvPercent = 2 	# /usr/include/X11/extensions/XI.h:202
DvPitch = 4 	# /usr/include/X11/extensions/XI.h:203
DvDuration = 8 	# /usr/include/X11/extensions/XI.h:204
DvLed = 16 	# /usr/include/X11/extensions/XI.h:205
DvLedMode = 32 	# /usr/include/X11/extensions/XI.h:206
DvKey = 64 	# /usr/include/X11/extensions/XI.h:207
DvAutoRepeatMode = 128 	# /usr/include/X11/extensions/XI.h:208
DvString = 1 	# /usr/include/X11/extensions/XI.h:210
DvInteger = 1 	# /usr/include/X11/extensions/XI.h:212
DeviceMode = 1 	# /usr/include/X11/extensions/XI.h:214
Relative = 0 	# /usr/include/X11/extensions/XI.h:215
Absolute = 1 	# /usr/include/X11/extensions/XI.h:216
ProximityState = 2 	# /usr/include/X11/extensions/XI.h:218
InProximity = 0 	# /usr/include/X11/extensions/XI.h:219
OutOfProximity = 2 	# /usr/include/X11/extensions/XI.h:220
AddToList = 0 	# /usr/include/X11/extensions/XI.h:222
DeleteFromList = 1 	# /usr/include/X11/extensions/XI.h:223
KeyClass = 0 	# /usr/include/X11/extensions/XI.h:225
ButtonClass = 1 	# /usr/include/X11/extensions/XI.h:226
ValuatorClass = 2 	# /usr/include/X11/extensions/XI.h:227
FeedbackClass = 3 	# /usr/include/X11/extensions/XI.h:228
ProximityClass = 4 	# /usr/include/X11/extensions/XI.h:229
FocusClass = 5 	# /usr/include/X11/extensions/XI.h:230
OtherClass = 6 	# /usr/include/X11/extensions/XI.h:231
KbdFeedbackClass = 0 	# /usr/include/X11/extensions/XI.h:233
PtrFeedbackClass = 1 	# /usr/include/X11/extensions/XI.h:234
StringFeedbackClass = 2 	# /usr/include/X11/extensions/XI.h:235
IntegerFeedbackClass = 3 	# /usr/include/X11/extensions/XI.h:236
LedFeedbackClass = 4 	# /usr/include/X11/extensions/XI.h:237
BellFeedbackClass = 5 	# /usr/include/X11/extensions/XI.h:238
_devicePointerMotionHint = 0 	# /usr/include/X11/extensions/XI.h:240
_deviceButton1Motion = 1 	# /usr/include/X11/extensions/XI.h:241
_deviceButton2Motion = 2 	# /usr/include/X11/extensions/XI.h:242
_deviceButton3Motion = 3 	# /usr/include/X11/extensions/XI.h:243
_deviceButton4Motion = 4 	# /usr/include/X11/extensions/XI.h:244
_deviceButton5Motion = 5 	# /usr/include/X11/extensions/XI.h:245
_deviceButtonMotion = 6 	# /usr/include/X11/extensions/XI.h:246
_deviceButtonGrab = 7 	# /usr/include/X11/extensions/XI.h:247
_deviceOwnerGrabButton = 8 	# /usr/include/X11/extensions/XI.h:248
_noExtensionEvent = 9 	# /usr/include/X11/extensions/XI.h:249
_devicePresence = 0 	# /usr/include/X11/extensions/XI.h:251
DeviceAdded = 0 	# /usr/include/X11/extensions/XI.h:253
DeviceRemoved = 1 	# /usr/include/X11/extensions/XI.h:254
DeviceEnabled = 2 	# /usr/include/X11/extensions/XI.h:255
DeviceDisabled = 3 	# /usr/include/X11/extensions/XI.h:256
DeviceUnrecoverable = 4 	# /usr/include/X11/extensions/XI.h:257
XI_BadDevice = 0 	# /usr/include/X11/extensions/XI.h:259
XI_BadEvent = 1 	# /usr/include/X11/extensions/XI.h:260
XI_BadMode = 2 	# /usr/include/X11/extensions/XI.h:261
XI_DeviceBusy = 3 	# /usr/include/X11/extensions/XI.h:262
XI_BadClass = 4 	# /usr/include/X11/extensions/XI.h:263
XEventClass = c_ulong 	# /usr/include/X11/extensions/XI.h:272
class struct_anon_93(Structure):
    __slots__ = [
        'present',
        'major_version',
        'minor_version',
    ]
struct_anon_93._fields_ = [
    ('present', c_int),
    ('major_version', c_short),
    ('minor_version', c_short),
]

XExtensionVersion = struct_anon_93 	# /usr/include/X11/extensions/XI.h:285
_deviceKeyPress = 0 	# /usr/include/X11/extensions/XInput.h:4902
_deviceKeyRelease = 1 	# /usr/include/X11/extensions/XInput.h:4903
_deviceButtonPress = 0 	# /usr/include/X11/extensions/XInput.h:4905
_deviceButtonRelease = 1 	# /usr/include/X11/extensions/XInput.h:4906
_deviceMotionNotify = 0 	# /usr/include/X11/extensions/XInput.h:4908
_deviceFocusIn = 0 	# /usr/include/X11/extensions/XInput.h:4910
_deviceFocusOut = 1 	# /usr/include/X11/extensions/XInput.h:4911
_proximityIn = 0 	# /usr/include/X11/extensions/XInput.h:4913
_proximityOut = 1 	# /usr/include/X11/extensions/XInput.h:4914
_deviceStateNotify = 0 	# /usr/include/X11/extensions/XInput.h:4916
_deviceMappingNotify = 1 	# /usr/include/X11/extensions/XInput.h:4917
_changeDeviceNotify = 2 	# /usr/include/X11/extensions/XInput.h:4918
class struct_anon_94(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'root',
        'subwindow',
        'time',
        'x',
        'y',
        'x_root',
        'y_root',
        'state',
        'keycode',
        'same_screen',
        'device_state',
        'axes_count',
        'first_axis',
        'axis_data',
    ]
Display = pyglet.libs.x11.xlib.Display
Window = pyglet.libs.x11.xlib.Window
XID = pyglet.libs.x11.xlib.XID
Time = pyglet.libs.x11.xlib.Time
struct_anon_94._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('root', Window),
    ('subwindow', Window),
    ('time', Time),
    ('x', c_int),
    ('y', c_int),
    ('x_root', c_int),
    ('y_root', c_int),
    ('state', c_uint),
    ('keycode', c_uint),
    ('same_screen', c_int),
    ('device_state', c_uint),
    ('axes_count', c_ubyte),
    ('first_axis', c_ubyte),
    ('axis_data', c_int * 6),
]

XDeviceKeyEvent = struct_anon_94 	# /usr/include/X11/extensions/XInput.h:5043
XDeviceKeyPressedEvent = XDeviceKeyEvent 	# /usr/include/X11/extensions/XInput.h:5045
XDeviceKeyReleasedEvent = XDeviceKeyEvent 	# /usr/include/X11/extensions/XInput.h:5046
class struct_anon_95(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'root',
        'subwindow',
        'time',
        'x',
        'y',
        'x_root',
        'y_root',
        'state',
        'button',
        'same_screen',
        'device_state',
        'axes_count',
        'first_axis',
        'axis_data',
    ]
struct_anon_95._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('root', Window),
    ('subwindow', Window),
    ('time', Time),
    ('x', c_int),
    ('y', c_int),
    ('x_root', c_int),
    ('y_root', c_int),
    ('state', c_uint),
    ('button', c_uint),
    ('same_screen', c_int),
    ('device_state', c_uint),
    ('axes_count', c_ubyte),
    ('first_axis', c_ubyte),
    ('axis_data', c_int * 6),
]

XDeviceButtonEvent = struct_anon_95 	# /usr/include/X11/extensions/XInput.h:5075
XDeviceButtonPressedEvent = XDeviceButtonEvent 	# /usr/include/X11/extensions/XInput.h:5077
XDeviceButtonReleasedEvent = XDeviceButtonEvent 	# /usr/include/X11/extensions/XInput.h:5078
class struct_anon_96(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'root',
        'subwindow',
        'time',
        'x',
        'y',
        'x_root',
        'y_root',
        'state',
        'is_hint',
        'same_screen',
        'device_state',
        'axes_count',
        'first_axis',
        'axis_data',
    ]
struct_anon_96._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('root', Window),
    ('subwindow', Window),
    ('time', Time),
    ('x', c_int),
    ('y', c_int),
    ('x_root', c_int),
    ('y_root', c_int),
    ('state', c_uint),
    ('is_hint', c_char),
    ('same_screen', c_int),
    ('device_state', c_uint),
    ('axes_count', c_ubyte),
    ('first_axis', c_ubyte),
    ('axis_data', c_int * 6),
]

XDeviceMotionEvent = struct_anon_96 	# /usr/include/X11/extensions/XInput.h:5108
class struct_anon_97(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'mode',
        'detail',
        'time',
    ]
struct_anon_97._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('mode', c_int),
    ('detail', c_int),
    ('time', Time),
]

XDeviceFocusChangeEvent = struct_anon_97 	# /usr/include/X11/extensions/XInput.h:5133
XDeviceFocusInEvent = XDeviceFocusChangeEvent 	# /usr/include/X11/extensions/XInput.h:5135
XDeviceFocusOutEvent = XDeviceFocusChangeEvent 	# /usr/include/X11/extensions/XInput.h:5136
class struct_anon_98(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'root',
        'subwindow',
        'time',
        'x',
        'y',
        'x_root',
        'y_root',
        'state',
        'same_screen',
        'device_state',
        'axes_count',
        'first_axis',
        'axis_data',
    ]
struct_anon_98._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('root', Window),
    ('subwindow', Window),
    ('time', Time),
    ('x', c_int),
    ('y', c_int),
    ('x_root', c_int),
    ('y_root', c_int),
    ('state', c_uint),
    ('same_screen', c_int),
    ('device_state', c_uint),
    ('axes_count', c_ubyte),
    ('first_axis', c_ubyte),
    ('axis_data', c_int * 6),
]

XProximityNotifyEvent = struct_anon_98 	# /usr/include/X11/extensions/XInput.h:5164
XProximityInEvent = XProximityNotifyEvent 	# /usr/include/X11/extensions/XInput.h:5165
XProximityOutEvent = XProximityNotifyEvent 	# /usr/include/X11/extensions/XInput.h:5166
class struct_anon_99(Structure):
    __slots__ = [
        'class',
        'length',
    ]
struct_anon_99._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
]

XInputClass = struct_anon_99 	# /usr/include/X11/extensions/XInput.h:5183
class struct_anon_100(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'time',
        'num_classes',
        'data',
    ]
struct_anon_100._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('time', Time),
    ('num_classes', c_int),
    ('data', c_char * 64),
]

XDeviceStateNotifyEvent = struct_anon_100 	# /usr/include/X11/extensions/XInput.h:5195
class struct_anon_101(Structure):
    __slots__ = [
        'class',
        'length',
        'num_valuators',
        'mode',
        'valuators',
    ]
struct_anon_101._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
    ('num_valuators', c_ubyte),
    ('mode', c_ubyte),
    ('valuators', c_int * 6),
]

XValuatorStatus = struct_anon_101 	# /usr/include/X11/extensions/XInput.h:5207
class struct_anon_102(Structure):
    __slots__ = [
        'class',
        'length',
        'num_keys',
        'keys',
    ]
struct_anon_102._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
    ('num_keys', c_short),
    ('keys', c_char * 32),
]

XKeyStatus = struct_anon_102 	# /usr/include/X11/extensions/XInput.h:5218
class struct_anon_103(Structure):
    __slots__ = [
        'class',
        'length',
        'num_buttons',
        'buttons',
    ]
struct_anon_103._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
    ('num_buttons', c_short),
    ('buttons', c_char * 32),
]

XButtonStatus = struct_anon_103 	# /usr/include/X11/extensions/XInput.h:5229
class struct_anon_104(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'time',
        'request',
        'first_keycode',
        'count',
    ]
struct_anon_104._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('time', Time),
    ('request', c_int),
    ('first_keycode', c_int),
    ('count', c_int),
]

XDeviceMappingEvent = struct_anon_104 	# /usr/include/X11/extensions/XInput.h:5250
class struct_anon_105(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'deviceid',
        'time',
        'request',
    ]
struct_anon_105._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('deviceid', XID),
    ('time', Time),
    ('request', c_int),
]

XChangeDeviceNotifyEvent = struct_anon_105 	# /usr/include/X11/extensions/XInput.h:5268
class struct_anon_106(Structure):
    __slots__ = [
        'type',
        'serial',
        'send_event',
        'display',
        'window',
        'time',
        'devchange',
        'deviceid',
        'control',
    ]
struct_anon_106._fields_ = [
    ('type', c_int),
    ('serial', c_ulong),
    ('send_event', c_int),
    ('display', POINTER(Display)),
    ('window', Window),
    ('time', Time),
    ('devchange', c_int),
    ('deviceid', XID),
    ('control', XID),
]

XDevicePresenceNotifyEvent = struct_anon_106 	# /usr/include/X11/extensions/XInput.h:5293
class struct_anon_107(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
    ]
struct_anon_107._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
]

XFeedbackState = struct_anon_107 	# /usr/include/X11/extensions/XInput.h:5311
class struct_anon_108(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'click',
        'percent',
        'pitch',
        'duration',
        'led_mask',
        'global_auto_repeat',
        'auto_repeats',
    ]
struct_anon_108._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('click', c_int),
    ('percent', c_int),
    ('pitch', c_int),
    ('duration', c_int),
    ('led_mask', c_int),
    ('global_auto_repeat', c_int),
    ('auto_repeats', c_char * 32),
]

XKbdFeedbackState = struct_anon_108 	# /usr/include/X11/extensions/XInput.h:5328
class struct_anon_109(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'accelNum',
        'accelDenom',
        'threshold',
    ]
struct_anon_109._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('accelNum', c_int),
    ('accelDenom', c_int),
    ('threshold', c_int),
]

XPtrFeedbackState = struct_anon_109 	# /usr/include/X11/extensions/XInput.h:5341
class struct_anon_110(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'resolution',
        'minVal',
        'maxVal',
    ]
struct_anon_110._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('resolution', c_int),
    ('minVal', c_int),
    ('maxVal', c_int),
]

XIntegerFeedbackState = struct_anon_110 	# /usr/include/X11/extensions/XInput.h:5354
class struct_anon_111(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'max_symbols',
        'num_syms_supported',
        'syms_supported',
    ]
KeySym = pyglet.libs.x11.xlib.KeySym
struct_anon_111._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('max_symbols', c_int),
    ('num_syms_supported', c_int),
    ('syms_supported', POINTER(KeySym)),
]

XStringFeedbackState = struct_anon_111 	# /usr/include/X11/extensions/XInput.h:5367
class struct_anon_112(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'percent',
        'pitch',
        'duration',
    ]
struct_anon_112._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('percent', c_int),
    ('pitch', c_int),
    ('duration', c_int),
]

XBellFeedbackState = struct_anon_112 	# /usr/include/X11/extensions/XInput.h:5380
class struct_anon_113(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'led_values',
        'led_mask',
    ]
struct_anon_113._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('led_values', c_int),
    ('led_mask', c_int),
]

XLedFeedbackState = struct_anon_113 	# /usr/include/X11/extensions/XInput.h:5392
class struct_anon_114(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
    ]
struct_anon_114._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
]

XFeedbackControl = struct_anon_114 	# /usr/include/X11/extensions/XInput.h:5402
class struct_anon_115(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'accelNum',
        'accelDenom',
        'threshold',
    ]
struct_anon_115._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('accelNum', c_int),
    ('accelDenom', c_int),
    ('threshold', c_int),
]

XPtrFeedbackControl = struct_anon_115 	# /usr/include/X11/extensions/XInput.h:5415
class struct_anon_116(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'click',
        'percent',
        'pitch',
        'duration',
        'led_mask',
        'led_value',
        'key',
        'auto_repeat_mode',
    ]
struct_anon_116._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('click', c_int),
    ('percent', c_int),
    ('pitch', c_int),
    ('duration', c_int),
    ('led_mask', c_int),
    ('led_value', c_int),
    ('key', c_int),
    ('auto_repeat_mode', c_int),
]

XKbdFeedbackControl = struct_anon_116 	# /usr/include/X11/extensions/XInput.h:5433
class struct_anon_117(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'num_keysyms',
        'syms_to_display',
    ]
struct_anon_117._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('num_keysyms', c_int),
    ('syms_to_display', POINTER(KeySym)),
]

XStringFeedbackControl = struct_anon_117 	# /usr/include/X11/extensions/XInput.h:5445
class struct_anon_118(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'int_to_display',
    ]
struct_anon_118._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('int_to_display', c_int),
]

XIntegerFeedbackControl = struct_anon_118 	# /usr/include/X11/extensions/XInput.h:5456
class struct_anon_119(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'percent',
        'pitch',
        'duration',
    ]
struct_anon_119._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('percent', c_int),
    ('pitch', c_int),
    ('duration', c_int),
]

XBellFeedbackControl = struct_anon_119 	# /usr/include/X11/extensions/XInput.h:5469
class struct_anon_120(Structure):
    __slots__ = [
        'class',
        'length',
        'id',
        'led_mask',
        'led_values',
    ]
struct_anon_120._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('id', XID),
    ('led_mask', c_int),
    ('led_values', c_int),
]

XLedFeedbackControl = struct_anon_120 	# /usr/include/X11/extensions/XInput.h:5481
class struct_anon_121(Structure):
    __slots__ = [
        'control',
        'length',
    ]
struct_anon_121._fields_ = [
    ('control', XID),
    ('length', c_int),
]

XDeviceControl = struct_anon_121 	# /usr/include/X11/extensions/XInput.h:5492
class struct_anon_122(Structure):
    __slots__ = [
        'control',
        'length',
        'first_valuator',
        'num_valuators',
        'resolutions',
    ]
struct_anon_122._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('first_valuator', c_int),
    ('num_valuators', c_int),
    ('resolutions', POINTER(c_int)),
]

XDeviceResolutionControl = struct_anon_122 	# /usr/include/X11/extensions/XInput.h:5500
class struct_anon_123(Structure):
    __slots__ = [
        'control',
        'length',
        'num_valuators',
        'resolutions',
        'min_resolutions',
        'max_resolutions',
    ]
struct_anon_123._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('num_valuators', c_int),
    ('resolutions', POINTER(c_int)),
    ('min_resolutions', POINTER(c_int)),
    ('max_resolutions', POINTER(c_int)),
]

XDeviceResolutionState = struct_anon_123 	# /usr/include/X11/extensions/XInput.h:5509
class struct_anon_124(Structure):
    __slots__ = [
        'control',
        'length',
        'min_x',
        'max_x',
        'min_y',
        'max_y',
        'flip_x',
        'flip_y',
        'rotation',
        'button_threshold',
    ]
struct_anon_124._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('min_x', c_int),
    ('max_x', c_int),
    ('min_y', c_int),
    ('max_y', c_int),
    ('flip_x', c_int),
    ('flip_y', c_int),
    ('rotation', c_int),
    ('button_threshold', c_int),
]

XDeviceAbsCalibControl = struct_anon_124 	# /usr/include/X11/extensions/XInput.h:5522
class struct_anon_125(Structure):
    __slots__ = [
        'control',
        'length',
        'min_x',
        'max_x',
        'min_y',
        'max_y',
        'flip_x',
        'flip_y',
        'rotation',
        'button_threshold',
    ]
struct_anon_125._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('min_x', c_int),
    ('max_x', c_int),
    ('min_y', c_int),
    ('max_y', c_int),
    ('flip_x', c_int),
    ('flip_y', c_int),
    ('rotation', c_int),
    ('button_threshold', c_int),
]

XDeviceAbsCalibState = struct_anon_125 	# /usr/include/X11/extensions/XInput.h:5522
class struct_anon_126(Structure):
    __slots__ = [
        'control',
        'length',
        'offset_x',
        'offset_y',
        'width',
        'height',
        'screen',
        'following',
    ]
struct_anon_126._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('offset_x', c_int),
    ('offset_y', c_int),
    ('width', c_int),
    ('height', c_int),
    ('screen', c_int),
    ('following', XID),
]

XDeviceAbsAreaControl = struct_anon_126 	# /usr/include/X11/extensions/XInput.h:5533
class struct_anon_127(Structure):
    __slots__ = [
        'control',
        'length',
        'offset_x',
        'offset_y',
        'width',
        'height',
        'screen',
        'following',
    ]
struct_anon_127._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('offset_x', c_int),
    ('offset_y', c_int),
    ('width', c_int),
    ('height', c_int),
    ('screen', c_int),
    ('following', XID),
]

XDeviceAbsAreaState = struct_anon_127 	# /usr/include/X11/extensions/XInput.h:5533
class struct_anon_128(Structure):
    __slots__ = [
        'control',
        'length',
        'status',
    ]
struct_anon_128._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('status', c_int),
]

XDeviceCoreControl = struct_anon_128 	# /usr/include/X11/extensions/XInput.h:5539
class struct_anon_129(Structure):
    __slots__ = [
        'control',
        'length',
        'status',
        'iscore',
    ]
struct_anon_129._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('status', c_int),
    ('iscore', c_int),
]

XDeviceCoreState = struct_anon_129 	# /usr/include/X11/extensions/XInput.h:5546
class struct_anon_130(Structure):
    __slots__ = [
        'control',
        'length',
        'enable',
    ]
struct_anon_130._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('enable', c_int),
]

XDeviceEnableControl = struct_anon_130 	# /usr/include/X11/extensions/XInput.h:5552
class struct_anon_131(Structure):
    __slots__ = [
        'control',
        'length',
        'enable',
    ]
struct_anon_131._fields_ = [
    ('control', XID),
    ('length', c_int),
    ('enable', c_int),
]

XDeviceEnableState = struct_anon_131 	# /usr/include/X11/extensions/XInput.h:5552
class struct__XAnyClassinfo(Structure):
    __slots__ = [
    ]
struct__XAnyClassinfo._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XAnyClassinfo(Structure):
    __slots__ = [
    ]
struct__XAnyClassinfo._fields_ = [
    ('_opaque_struct', c_int)
]

XAnyClassPtr = POINTER(struct__XAnyClassinfo) 	# /usr/include/X11/extensions/XInput.h:5564
class struct__XAnyClassinfo(Structure):
    __slots__ = [
        'class',
        'length',
    ]
struct__XAnyClassinfo._fields_ = [
    ('class', XID),
    ('length', c_int),
]

XAnyClassInfo = struct__XAnyClassinfo 	# /usr/include/X11/extensions/XInput.h:5573
class struct__XDeviceInfo(Structure):
    __slots__ = [
    ]
struct__XDeviceInfo._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XDeviceInfo(Structure):
    __slots__ = [
    ]
struct__XDeviceInfo._fields_ = [
    ('_opaque_struct', c_int)
]

XDeviceInfoPtr = POINTER(struct__XDeviceInfo) 	# /usr/include/X11/extensions/XInput.h:5575
class struct__XDeviceInfo(Structure):
    __slots__ = [
        'id',
        'type',
        'name',
        'num_classes',
        'use',
        'inputclassinfo',
    ]
Atom = pyglet.libs.x11.xlib.Atom
struct__XDeviceInfo._fields_ = [
    ('id', XID),
    ('type', Atom),
    ('name', c_char_p),
    ('num_classes', c_int),
    ('use', c_int),
    ('inputclassinfo', XAnyClassPtr),
]

XDeviceInfo = struct__XDeviceInfo 	# /usr/include/X11/extensions/XInput.h:5585
class struct__XKeyInfo(Structure):
    __slots__ = [
    ]
struct__XKeyInfo._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XKeyInfo(Structure):
    __slots__ = [
    ]
struct__XKeyInfo._fields_ = [
    ('_opaque_struct', c_int)
]

XKeyInfoPtr = POINTER(struct__XKeyInfo) 	# /usr/include/X11/extensions/XInput.h:5587
class struct__XKeyInfo(Structure):
    __slots__ = [
        'class',
        'length',
        'min_keycode',
        'max_keycode',
        'num_keys',
    ]
struct__XKeyInfo._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('min_keycode', c_ushort),
    ('max_keycode', c_ushort),
    ('num_keys', c_ushort),
]

XKeyInfo = struct__XKeyInfo 	# /usr/include/X11/extensions/XInput.h:5600
class struct__XButtonInfo(Structure):
    __slots__ = [
    ]
struct__XButtonInfo._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XButtonInfo(Structure):
    __slots__ = [
    ]
struct__XButtonInfo._fields_ = [
    ('_opaque_struct', c_int)
]

XButtonInfoPtr = POINTER(struct__XButtonInfo) 	# /usr/include/X11/extensions/XInput.h:5602
class struct__XButtonInfo(Structure):
    __slots__ = [
        'class',
        'length',
        'num_buttons',
    ]
struct__XButtonInfo._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('num_buttons', c_short),
]

XButtonInfo = struct__XButtonInfo 	# /usr/include/X11/extensions/XInput.h:5612
class struct__XAxisInfo(Structure):
    __slots__ = [
    ]
struct__XAxisInfo._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XAxisInfo(Structure):
    __slots__ = [
    ]
struct__XAxisInfo._fields_ = [
    ('_opaque_struct', c_int)
]

XAxisInfoPtr = POINTER(struct__XAxisInfo) 	# /usr/include/X11/extensions/XInput.h:5614
class struct__XAxisInfo(Structure):
    __slots__ = [
        'resolution',
        'min_value',
        'max_value',
    ]
struct__XAxisInfo._fields_ = [
    ('resolution', c_int),
    ('min_value', c_int),
    ('max_value', c_int),
]

XAxisInfo = struct__XAxisInfo 	# /usr/include/X11/extensions/XInput.h:5620
class struct__XValuatorInfo(Structure):
    __slots__ = [
    ]
struct__XValuatorInfo._fields_ = [
    ('_opaque_struct', c_int)
]

class struct__XValuatorInfo(Structure):
    __slots__ = [
    ]
struct__XValuatorInfo._fields_ = [
    ('_opaque_struct', c_int)
]

XValuatorInfoPtr = POINTER(struct__XValuatorInfo) 	# /usr/include/X11/extensions/XInput.h:5622
class struct__XValuatorInfo(Structure):
    __slots__ = [
        'class',
        'length',
        'num_axes',
        'mode',
        'motion_buffer',
        'axes',
    ]
struct__XValuatorInfo._fields_ = [
    ('class', XID),
    ('length', c_int),
    ('num_axes', c_ubyte),
    ('mode', c_ubyte),
    ('motion_buffer', c_ulong),
    ('axes', XAxisInfoPtr),
]

XValuatorInfo = struct__XValuatorInfo 	# /usr/include/X11/extensions/XInput.h:5636
class struct_anon_132(Structure):
    __slots__ = [
        'input_class',
        'event_type_base',
    ]
struct_anon_132._fields_ = [
    ('input_class', c_ubyte),
    ('event_type_base', c_ubyte),
]

XInputClassInfo = struct_anon_132 	# /usr/include/X11/extensions/XInput.h:5653
class struct_anon_133(Structure):
    __slots__ = [
        'device_id',
        'num_classes',
        'classes',
    ]
struct_anon_133._fields_ = [
    ('device_id', XID),
    ('num_classes', c_int),
    ('classes', POINTER(XInputClassInfo)),
]

XDevice = struct_anon_133 	# /usr/include/X11/extensions/XInput.h:5659
class struct_anon_134(Structure):
    __slots__ = [
        'event_type',
        'device',
    ]
struct_anon_134._fields_ = [
    ('event_type', XEventClass),
    ('device', XID),
]

XEventList = struct_anon_134 	# /usr/include/X11/extensions/XInput.h:5672
class struct_anon_135(Structure):
    __slots__ = [
        'time',
        'data',
    ]
struct_anon_135._fields_ = [
    ('time', Time),
    ('data', POINTER(c_int)),
]

XDeviceTimeCoord = struct_anon_135 	# /usr/include/X11/extensions/XInput.h:5685
class struct_anon_136(Structure):
    __slots__ = [
        'device_id',
        'num_classes',
        'data',
    ]
struct_anon_136._fields_ = [
    ('device_id', XID),
    ('num_classes', c_int),
    ('data', POINTER(XInputClass)),
]

XDeviceState = struct_anon_136 	# /usr/include/X11/extensions/XInput.h:5699
class struct_anon_137(Structure):
    __slots__ = [
        'class',
        'length',
        'num_valuators',
        'mode',
        'valuators',
    ]
struct_anon_137._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
    ('num_valuators', c_ubyte),
    ('mode', c_ubyte),
    ('valuators', POINTER(c_int)),
]

XValuatorState = struct_anon_137 	# /usr/include/X11/extensions/XInput.h:5722
class struct_anon_138(Structure):
    __slots__ = [
        'class',
        'length',
        'num_keys',
        'keys',
    ]
struct_anon_138._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
    ('num_keys', c_short),
    ('keys', c_char * 32),
]

XKeyState = struct_anon_138 	# /usr/include/X11/extensions/XInput.h:5733
class struct_anon_139(Structure):
    __slots__ = [
        'class',
        'length',
        'num_buttons',
        'buttons',
    ]
struct_anon_139._fields_ = [
    ('class', c_ubyte),
    ('length', c_ubyte),
    ('num_buttons', c_short),
    ('buttons', c_char * 32),
]

XButtonState = struct_anon_139 	# /usr/include/X11/extensions/XInput.h:5744
# /usr/include/X11/extensions/XInput.h:5754
XChangeKeyboardDevice = _lib.XChangeKeyboardDevice
XChangeKeyboardDevice.restype = c_int
XChangeKeyboardDevice.argtypes = [POINTER(Display), POINTER(XDevice)]

# /usr/include/X11/extensions/XInput.h:5759
XChangePointerDevice = _lib.XChangePointerDevice
XChangePointerDevice.restype = c_int
XChangePointerDevice.argtypes = [POINTER(Display), POINTER(XDevice), c_int, c_int]

# /usr/include/X11/extensions/XInput.h:5766
XGrabDevice = _lib.XGrabDevice
XGrabDevice.restype = c_int
XGrabDevice.argtypes = [POINTER(Display), POINTER(XDevice), Window, c_int, c_int, POINTER(XEventClass), c_int, c_int, Time]

# /usr/include/X11/extensions/XInput.h:5778
XUngrabDevice = _lib.XUngrabDevice
XUngrabDevice.restype = c_int
XUngrabDevice.argtypes = [POINTER(Display), POINTER(XDevice), Time]

# /usr/include/X11/extensions/XInput.h:5784
XGrabDeviceKey = _lib.XGrabDeviceKey
XGrabDeviceKey.restype = c_int
XGrabDeviceKey.argtypes = [POINTER(Display), POINTER(XDevice), c_uint, c_uint, POINTER(XDevice), Window, c_int, c_uint, POINTER(XEventClass), c_int, c_int]

# /usr/include/X11/extensions/XInput.h:5798
XUngrabDeviceKey = _lib.XUngrabDeviceKey
XUngrabDeviceKey.restype = c_int
XUngrabDeviceKey.argtypes = [POINTER(Display), POINTER(XDevice), c_uint, c_uint, POINTER(XDevice), Window]

# /usr/include/X11/extensions/XInput.h:5807
XGrabDeviceButton = _lib.XGrabDeviceButton
XGrabDeviceButton.restype = c_int
XGrabDeviceButton.argtypes = [POINTER(Display), POINTER(XDevice), c_uint, c_uint, POINTER(XDevice), Window, c_int, c_uint, POINTER(XEventClass), c_int, c_int]

# /usr/include/X11/extensions/XInput.h:5821
XUngrabDeviceButton = _lib.XUngrabDeviceButton
XUngrabDeviceButton.restype = c_int
XUngrabDeviceButton.argtypes = [POINTER(Display), POINTER(XDevice), c_uint, c_uint, POINTER(XDevice), Window]

# /usr/include/X11/extensions/XInput.h:5830
XAllowDeviceEvents = _lib.XAllowDeviceEvents
XAllowDeviceEvents.restype = c_int
XAllowDeviceEvents.argtypes = [POINTER(Display), POINTER(XDevice), c_int, Time]

# /usr/include/X11/extensions/XInput.h:5837
XGetDeviceFocus = _lib.XGetDeviceFocus
XGetDeviceFocus.restype = c_int
XGetDeviceFocus.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(Window), POINTER(c_int), POINTER(Time)]

# /usr/include/X11/extensions/XInput.h:5845
XSetDeviceFocus = _lib.XSetDeviceFocus
XSetDeviceFocus.restype = c_int
XSetDeviceFocus.argtypes = [POINTER(Display), POINTER(XDevice), Window, c_int, Time]

# /usr/include/X11/extensions/XInput.h:5853
XGetFeedbackControl = _lib.XGetFeedbackControl
XGetFeedbackControl.restype = POINTER(XFeedbackState)
XGetFeedbackControl.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(c_int)]

# /usr/include/X11/extensions/XInput.h:5859
XFreeFeedbackList = _lib.XFreeFeedbackList
XFreeFeedbackList.restype = None
XFreeFeedbackList.argtypes = [POINTER(XFeedbackState)]

# /usr/include/X11/extensions/XInput.h:5863
XChangeFeedbackControl = _lib.XChangeFeedbackControl
XChangeFeedbackControl.restype = c_int
XChangeFeedbackControl.argtypes = [POINTER(Display), POINTER(XDevice), c_ulong, POINTER(XFeedbackControl)]

# /usr/include/X11/extensions/XInput.h:5870
XDeviceBell = _lib.XDeviceBell
XDeviceBell.restype = c_int
XDeviceBell.argtypes = [POINTER(Display), POINTER(XDevice), XID, XID, c_int]

KeyCode = pyglet.libs.x11.xlib.KeyCode
# /usr/include/X11/extensions/XInput.h:5878
XGetDeviceKeyMapping = _lib.XGetDeviceKeyMapping
XGetDeviceKeyMapping.restype = POINTER(KeySym)
XGetDeviceKeyMapping.argtypes = [POINTER(Display), POINTER(XDevice), KeyCode, c_int, POINTER(c_int)]

# /usr/include/X11/extensions/XInput.h:5890
XChangeDeviceKeyMapping = _lib.XChangeDeviceKeyMapping
XChangeDeviceKeyMapping.restype = c_int
XChangeDeviceKeyMapping.argtypes = [POINTER(Display), POINTER(XDevice), c_int, c_int, POINTER(KeySym), c_int]

XModifierKeymap = pyglet.libs.x11.xlib.XModifierKeymap
# /usr/include/X11/extensions/XInput.h:5899
XGetDeviceModifierMapping = _lib.XGetDeviceModifierMapping
XGetDeviceModifierMapping.restype = POINTER(XModifierKeymap)
XGetDeviceModifierMapping.argtypes = [POINTER(Display), POINTER(XDevice)]

# /usr/include/X11/extensions/XInput.h:5904
XSetDeviceModifierMapping = _lib.XSetDeviceModifierMapping
XSetDeviceModifierMapping.restype = c_int
XSetDeviceModifierMapping.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(XModifierKeymap)]

# /usr/include/X11/extensions/XInput.h:5910
XSetDeviceButtonMapping = _lib.XSetDeviceButtonMapping
XSetDeviceButtonMapping.restype = c_int
XSetDeviceButtonMapping.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(c_ubyte), c_int]

# /usr/include/X11/extensions/XInput.h:5917
XGetDeviceButtonMapping = _lib.XGetDeviceButtonMapping
XGetDeviceButtonMapping.restype = c_int
XGetDeviceButtonMapping.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(c_ubyte), c_uint]

# /usr/include/X11/extensions/XInput.h:5924
XQueryDeviceState = _lib.XQueryDeviceState
XQueryDeviceState.restype = POINTER(XDeviceState)
XQueryDeviceState.argtypes = [POINTER(Display), POINTER(XDevice)]

# /usr/include/X11/extensions/XInput.h:5929
XFreeDeviceState = _lib.XFreeDeviceState
XFreeDeviceState.restype = None
XFreeDeviceState.argtypes = [POINTER(XDeviceState)]

# /usr/include/X11/extensions/XInput.h:5933
XGetExtensionVersion = _lib.XGetExtensionVersion
XGetExtensionVersion.restype = POINTER(XExtensionVersion)
XGetExtensionVersion.argtypes = [POINTER(Display), c_char_p]

# /usr/include/X11/extensions/XInput.h:5938
XListInputDevices = _lib.XListInputDevices
XListInputDevices.restype = POINTER(XDeviceInfo)
XListInputDevices.argtypes = [POINTER(Display), POINTER(c_int)]

# /usr/include/X11/extensions/XInput.h
XListDeviceProperties = _lib.XListDeviceProperties
XListDeviceProperties.restype = POINTER(Atom)
XListDeviceProperties.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(c_int)]

# /usr/include/X11/extensions/XInput.h
XGetDeviceProperty = _lib.XGetDeviceProperty
XGetDeviceProperty.restype = c_int
XGetDeviceProperty.argtypes = [POINTER(Display), POINTER(XDevice), Atom, c_long, c_long, c_bool, Atom, POINTER(Atom), POINTER(c_int), POINTER(c_ulong), POINTER(c_ulong), POINTER(c_char_p)]

# /usr/include/X11/extensions/XInput.h:5943
XFreeDeviceList = _lib.XFreeDeviceList
XFreeDeviceList.restype = None
XFreeDeviceList.argtypes = [POINTER(XDeviceInfo)]

# /usr/include/X11/extensions/XInput.h:5947
XOpenDevice = _lib.XOpenDevice
XOpenDevice.restype = POINTER(XDevice)
XOpenDevice.argtypes = [POINTER(Display), XID]

# /usr/include/X11/extensions/XInput.h:5952
XCloseDevice = _lib.XCloseDevice
XCloseDevice.restype = c_int
XCloseDevice.argtypes = [POINTER(Display), POINTER(XDevice)]

# /usr/include/X11/extensions/XInput.h:5957
XSetDeviceMode = _lib.XSetDeviceMode
XSetDeviceMode.restype = c_int
XSetDeviceMode.argtypes = [POINTER(Display), POINTER(XDevice), c_int]

# /usr/include/X11/extensions/XInput.h:5963
XSetDeviceValuators = _lib.XSetDeviceValuators
XSetDeviceValuators.restype = c_int
XSetDeviceValuators.argtypes = [POINTER(Display), POINTER(XDevice), POINTER(c_int), c_int, c_int]

# /usr/include/X11/extensions/XInput.h:5971
XGetDeviceControl = _lib.XGetDeviceControl
XGetDeviceControl.restype = POINTER(XDeviceControl)
XGetDeviceControl.argtypes = [POINTER(Display), POINTER(XDevice), c_int]

# /usr/include/X11/extensions/XInput.h:5977
XChangeDeviceControl = _lib.XChangeDeviceControl
XChangeDeviceControl.restype = c_int
XChangeDeviceControl.argtypes = [POINTER(Display), POINTER(XDevice), c_int, POINTER(XDeviceControl)]

# /usr/include/X11/extensions/XInput.h:5984
XSelectExtensionEvent = _lib.XSelectExtensionEvent
XSelectExtensionEvent.restype = c_int
XSelectExtensionEvent.argtypes = [POINTER(Display), Window, POINTER(XEventClass), c_int]

# /usr/include/X11/extensions/XInput.h:5991
XGetSelectedExtensionEvents = _lib.XGetSelectedExtensionEvents
XGetSelectedExtensionEvents.restype = c_int
XGetSelectedExtensionEvents.argtypes = [POINTER(Display), Window, POINTER(c_int), POINTER(POINTER(XEventClass)), POINTER(c_int), POINTER(POINTER(XEventClass))]

# /usr/include/X11/extensions/XInput.h:6000
XChangeDeviceDontPropagateList = _lib.XChangeDeviceDontPropagateList
XChangeDeviceDontPropagateList.restype = c_int
XChangeDeviceDontPropagateList.argtypes = [POINTER(Display), Window, c_int, POINTER(XEventClass), c_int]

# /usr/include/X11/extensions/XInput.h:6008
XGetDeviceDontPropagateList = _lib.XGetDeviceDontPropagateList
XGetDeviceDontPropagateList.restype = POINTER(XEventClass)
XGetDeviceDontPropagateList.argtypes = [POINTER(Display), Window, POINTER(c_int)]

XEvent = pyglet.libs.x11.xlib.XEvent
# /usr/include/X11/extensions/XInput.h:6014
XSendExtensionEvent = _lib.XSendExtensionEvent
XSendExtensionEvent.restype = c_int
XSendExtensionEvent.argtypes = [POINTER(Display), POINTER(XDevice), Window, c_int, c_int, POINTER(XEventClass), POINTER(XEvent)]

# /usr/include/X11/extensions/XInput.h:6024
XGetDeviceMotionEvents = _lib.XGetDeviceMotionEvents
XGetDeviceMotionEvents.restype = POINTER(XDeviceTimeCoord)
XGetDeviceMotionEvents.argtypes = [POINTER(Display), POINTER(XDevice), Time, Time, POINTER(c_int), POINTER(c_int), POINTER(c_int)]

# /usr/include/X11/extensions/XInput.h:6034
XFreeDeviceMotionEvents = _lib.XFreeDeviceMotionEvents
XFreeDeviceMotionEvents.restype = None
XFreeDeviceMotionEvents.argtypes = [POINTER(XDeviceTimeCoord)]

# /usr/include/X11/extensions/XInput.h:6038
XFreeDeviceControl = _lib.XFreeDeviceControl
XFreeDeviceControl.restype = None
XFreeDeviceControl.argtypes = [POINTER(XDeviceControl)]


__all__ = ['sz_xGetExtensionVersionReq', 'sz_xGetExtensionVersionReply',
'sz_xListInputDevicesReq', 'sz_xListInputDevicesReply', 'sz_xOpenDeviceReq',
'sz_xOpenDeviceReply', 'sz_xCloseDeviceReq', 'sz_xSetDeviceModeReq',
'sz_xSetDeviceModeReply', 'sz_xSelectExtensionEventReq',
'sz_xGetSelectedExtensionEventsReq', 'sz_xGetSelectedExtensionEventsReply',
'sz_xChangeDeviceDontPropagateListReq', 'sz_xGetDeviceDontPropagateListReq',
'sz_xGetDeviceDontPropagateListReply', 'sz_xGetDeviceMotionEventsReq',
'sz_xGetDeviceMotionEventsReply', 'sz_xChangeKeyboardDeviceReq',
'sz_xChangeKeyboardDeviceReply', 'sz_xChangePointerDeviceReq',
'sz_xChangePointerDeviceReply', 'sz_xGrabDeviceReq', 'sz_xGrabDeviceReply',
'sz_xUngrabDeviceReq', 'sz_xGrabDeviceKeyReq', 'sz_xGrabDeviceKeyReply',
'sz_xUngrabDeviceKeyReq', 'sz_xGrabDeviceButtonReq',
'sz_xGrabDeviceButtonReply', 'sz_xUngrabDeviceButtonReq',
'sz_xAllowDeviceEventsReq', 'sz_xGetDeviceFocusReq',
'sz_xGetDeviceFocusReply', 'sz_xSetDeviceFocusReq',
'sz_xGetFeedbackControlReq', 'sz_xGetFeedbackControlReply',
'sz_xChangeFeedbackControlReq', 'sz_xGetDeviceKeyMappingReq',
'sz_xGetDeviceKeyMappingReply', 'sz_xChangeDeviceKeyMappingReq',
'sz_xGetDeviceModifierMappingReq', 'sz_xSetDeviceModifierMappingReq',
'sz_xSetDeviceModifierMappingReply', 'sz_xGetDeviceButtonMappingReq',
'sz_xGetDeviceButtonMappingReply', 'sz_xSetDeviceButtonMappingReq',
'sz_xSetDeviceButtonMappingReply', 'sz_xQueryDeviceStateReq',
'sz_xQueryDeviceStateReply', 'sz_xSendExtensionEventReq', 'sz_xDeviceBellReq',
'sz_xSetDeviceValuatorsReq', 'sz_xSetDeviceValuatorsReply',
'sz_xGetDeviceControlReq', 'sz_xGetDeviceControlReply',
'sz_xChangeDeviceControlReq', 'sz_xChangeDeviceControlReply', 'Dont_Check',
'XInput_Initial_Release', 'XInput_Add_XDeviceBell',
'XInput_Add_XSetDeviceValuators', 'XInput_Add_XChangeDeviceControl',
'XInput_Add_DevicePresenceNotify', 'XI_Absent', 'XI_Present',
'XI_Initial_Release_Major', 'XI_Initial_Release_Minor',
'XI_Add_XDeviceBell_Major', 'XI_Add_XDeviceBell_Minor',
'XI_Add_XSetDeviceValuators_Major', 'XI_Add_XSetDeviceValuators_Minor',
'XI_Add_XChangeDeviceControl_Major', 'XI_Add_XChangeDeviceControl_Minor',
'XI_Add_DevicePresenceNotify_Major', 'XI_Add_DevicePresenceNotify_Minor',
'DEVICE_RESOLUTION', 'DEVICE_ABS_CALIB', 'DEVICE_CORE', 'DEVICE_ENABLE',
'DEVICE_ABS_AREA', 'NoSuchExtension', 'COUNT', 'CREATE', 'NewPointer',
'NewKeyboard', 'XPOINTER', 'XKEYBOARD', 'UseXKeyboard', 'IsXPointer',
'IsXKeyboard', 'IsXExtensionDevice', 'IsXExtensionKeyboard',
'IsXExtensionPointer', 'AsyncThisDevice', 'SyncThisDevice',
'ReplayThisDevice', 'AsyncOtherDevices', 'AsyncAll', 'SyncAll',
'FollowKeyboard', 'RevertToFollowKeyboard', 'DvAccelNum', 'DvAccelDenom',
'DvThreshold', 'DvKeyClickPercent', 'DvPercent', 'DvPitch', 'DvDuration',
'DvLed', 'DvLedMode', 'DvKey', 'DvAutoRepeatMode', 'DvString', 'DvInteger',
'DeviceMode', 'Relative', 'Absolute', 'ProximityState', 'InProximity',
'OutOfProximity', 'AddToList', 'DeleteFromList', 'KeyClass', 'ButtonClass',
'ValuatorClass', 'FeedbackClass', 'ProximityClass', 'FocusClass',
'OtherClass', 'KbdFeedbackClass', 'PtrFeedbackClass', 'StringFeedbackClass',
'IntegerFeedbackClass', 'LedFeedbackClass', 'BellFeedbackClass',
'_devicePointerMotionHint', '_deviceButton1Motion', '_deviceButton2Motion',
'_deviceButton3Motion', '_deviceButton4Motion', '_deviceButton5Motion',
'_deviceButtonMotion', '_deviceButtonGrab', '_deviceOwnerGrabButton',
'_noExtensionEvent', '_devicePresence', 'DeviceAdded', 'DeviceRemoved',
'DeviceEnabled', 'DeviceDisabled', 'DeviceUnrecoverable', 'XI_BadDevice',
'XI_BadEvent', 'XI_BadMode', 'XI_DeviceBusy', 'XI_BadClass', 'XEventClass',
'XExtensionVersion', '_deviceKeyPress', '_deviceKeyRelease',
'_deviceButtonPress', '_deviceButtonRelease', '_deviceMotionNotify',
'_deviceFocusIn', '_deviceFocusOut', '_proximityIn', '_proximityOut',
'_deviceStateNotify', '_deviceMappingNotify', '_changeDeviceNotify',
'XDeviceKeyEvent', 'XDeviceKeyPressedEvent', 'XDeviceKeyReleasedEvent',
'XDeviceButtonEvent', 'XDeviceButtonPressedEvent',
'XDeviceButtonReleasedEvent', 'XDeviceMotionEvent', 'XDeviceFocusChangeEvent',
'XDeviceFocusInEvent', 'XDeviceFocusOutEvent', 'XProximityNotifyEvent',
'XProximityInEvent', 'XProximityOutEvent', 'XInputClass',
'XDeviceStateNotifyEvent', 'XValuatorStatus', 'XKeyStatus', 'XButtonStatus',
'XDeviceMappingEvent', 'XChangeDeviceNotifyEvent',
'XDevicePresenceNotifyEvent', 'XFeedbackState', 'XKbdFeedbackState',
'XPtrFeedbackState', 'XIntegerFeedbackState', 'XStringFeedbackState',
'XBellFeedbackState', 'XLedFeedbackState', 'XFeedbackControl',
'XPtrFeedbackControl', 'XKbdFeedbackControl', 'XStringFeedbackControl',
'XIntegerFeedbackControl', 'XBellFeedbackControl', 'XLedFeedbackControl',
'XDeviceControl', 'XDeviceResolutionControl', 'XDeviceResolutionState',
'XDeviceAbsCalibControl', 'XDeviceAbsCalibState', 'XDeviceAbsAreaControl',
'XDeviceAbsAreaState', 'XDeviceCoreControl', 'XDeviceCoreState',
'XDeviceEnableControl', 'XDeviceEnableState', 'XAnyClassPtr', 'XAnyClassInfo',
'XDeviceInfoPtr', 'XDeviceInfo', 'XKeyInfoPtr', 'XKeyInfo', 'XButtonInfoPtr',
'XButtonInfo', 'XAxisInfoPtr', 'XAxisInfo', 'XValuatorInfoPtr',
'XValuatorInfo', 'XInputClassInfo', 'XDevice', 'XEventList',
'XDeviceTimeCoord', 'XDeviceState', 'XValuatorState', 'XKeyState',
'XButtonState', 'XChangeKeyboardDevice', 'XChangePointerDevice',
'XGrabDevice', 'XUngrabDevice', 'XGrabDeviceKey', 'XUngrabDeviceKey',
'XGrabDeviceButton', 'XUngrabDeviceButton', 'XAllowDeviceEvents',
'XGetDeviceFocus', 'XSetDeviceFocus', 'XGetFeedbackControl',
'XFreeFeedbackList', 'XChangeFeedbackControl', 'XDeviceBell',
'XGetDeviceKeyMapping', 'XChangeDeviceKeyMapping',
'XGetDeviceModifierMapping', 'XSetDeviceModifierMapping',
'XSetDeviceButtonMapping', 'XGetDeviceButtonMapping', 'XQueryDeviceState',
'XFreeDeviceState', 'XGetExtensionVersion', 'XListInputDevices', 
'XListDeviceProperties', 'XGetDeviceProperty', 'XFreeDeviceList', 
'XOpenDevice', 'XCloseDevice', 'XSetDeviceMode',
'XSetDeviceValuators', 'XGetDeviceControl', 'XChangeDeviceControl',
'XSelectExtensionEvent', 'XGetSelectedExtensionEvents',
'XChangeDeviceDontPropagateList', 'XGetDeviceDontPropagateList',
'XSendExtensionEvent', 'XGetDeviceMotionEvents', 'XFreeDeviceMotionEvents',
'XFreeDeviceControl']
